import morsaik as kdi
import numpy as np

from numpy.testing import assert_equal

def test_MotifTrajectoryEnsemble():
    alphabet = ['a','b','c']
    motiflength = 5
    unit = kdi.make_unit('mol')/kdi.make_unit('L')
    times = kdi.TimesVector(np.arange(3), 'ms')

    motif_categories = kdi.domains._return_motif_categories(motiflength=motiflength)
    motif_number_dct = kdi._create_empty_motif_vector_dct(motiflength,alphabet=alphabet)

    motif_number_vector = kdi.MotifVector(motiflength,alphabet,unit)
    motif_number_vector = motif_number_vector(motif_number_dct)
    motif_number_vectors = [motif_number_vector,]*times.shape[0]
    mt = kdi.MotifTrajectory(motif_number_vectors,times)

    mts = [mt,mt]

    mte = kdi.MotifTrajectoryEnsemble(mts)

    assert_equal(mte.trajectories[0].times.val,mt.times.val)
    assert_equal(mte.trajectories[0].times.domain,mt.times.domain)
    assert_equal(mte.trajectories[1].times.val,mt.times.val)
    assert_equal(mte.trajectories[1].times.domain,mt.times.domain)
    assert_equal(mte.alphabet, mt.alphabet)
    assert_equal(mte.number_of_letters, mt.number_of_letters)
    assert(mte.motiflength==mt.motiflength)
    assert(mte.unit==mt.unit)
    for trajectory in mte.trajectories:
        for ii in range(len(trajectory.times.val)):
            assert_equal(trajectory.motifs.val,
                    mt.motifs.val
                    )


    print('mt.times')
    print(mt.times)
    assert(kdi.isinstance_motiftrajectoryensemble(mte))
    assert(kdi.are_compatible_motif_trajectory_ensembles(mte,mte))

def test_save_and_load_MotifTrajectoryEnsemble():
    alphabet = ['a','b','c']
    motiflengths = [2,3,4,5,6,10]
    unit = kdi.make_unit('mol')/kdi.make_unit('L')
    times = np.arange(3)

    for motiflength in motiflengths:
        strand_trajectory_id='9999_99_99__99_99_99'
        param_file_no = 0
        archive_path = kdi.utils.create_trajectory_ensemble_path(
            strand_trajectory_id=strand_trajectory_id,
            param_file_no = param_file_no,
            motiflength=motiflength
        )

        motif_trajectory = kdi.get.strand_motifs_trajectory(motiflength, strand_trajectory_id)
        motif_trajectories = [motif_trajectory,]*2
        mte = kdi.MotifTrajectoryEnsemble(motif_trajectories)

        kdi.save_motif_trajectory_ensemble(archive_path, mte)
        actual_mte = kdi.load_motif_trajectory_ensemble(archive_path)
        assert(mte.motiflength==actual_mte.motiflength)
        assert(mte.unit==actual_mte.unit)
        assert(mte.alphabet==actual_mte.alphabet)
        for trajectory_index in range(len(actual_mte.trajectories)):
            actual = kdi._motif_trajectory_as_array(actual_mte.trajectories[trajectory_index])
            assert_equal(actual,
                    kdi._motif_trajectory_as_array(motif_trajectories[trajectory_index])
                    )



from ..obj.units import make_unit

def config(config_yaml_path : str) -> dict:
    """
    reads yml file and returns it as dictionary

    Parameters:
    -----------
    config_yaml_path : string
        path of the yml-file

    Returns:
    --------
    config_dct : dict
    """
    import yaml
    with open(config_yaml_path) as f:
        config_dct = yaml.safe_load(f)
    print("Read config yaml.")
    return config_dct

def symbol_config(
        symbol : str = None,
        unitformat : bool = False
    ) -> dict:
    if symbol is None:
        return config('./config/symbols.yml')
    else:
        if unitformat:
            return make_unit(config('./config/symbols.yml')[symbol][1])
        else:
            return config('./config/symbols.yml')[symbol]



import jax
import jax.numpy as jnp
from typing import Callable

import nifty8.re as jft

from .collisions import collisions_from_motif_concentration_trajectory_array_and_collision_rate_constants_array

def motif_production_rates_array_from_motif_production_rate_constants_array_and_motif_concentrations_array(
        motif_production_rate_constants : jax.Array,
        motif_logconcentrations_trajectory : jax.Array,
        motif_concentration_trajectory_times_array : jax.Array = None,
        ):
    """
    calculates motif production rates from
    motif production rate constants and motif concentrations
    Parameters:
    -----------
    motif_production_rate_constants : jax.Array
    motif_concentration_trajectories : jax.Array

    Returns:
    --------
    motif_production_rates : jax.Array
    """
    motiflength = len(motif_logconcentrations_trajectory[0].shape)
    nol = motif_logconcentrations_trajectory[0].shape[0]-1
    mpr_shape = (nol+1,nol,nol,nol+1,nol+1,nol,nol,nol+1)
    if motiflength == 4:
        # \lambda_{ijk} c_i c_j c_k
        # integrate collision over time
        exposure = collisions_from_motif_concentration_trajectory_array_and_collision_rate_constants_array(
                motif_logconcentrations_trajectory,
                motif_concentration_trajectory_times_array=motif_concentration_trajectory_times_array,
                concentrations_are_logarithmised = True
                )
        motif_production_rates = motif_production_rate_constants.flatten()*exposure
        return motif_production_rates.reshape(mpr_shape)
    else:
        raise NotImplementedError("only implemented for motiflength 4")

def motif_production_rates_array_from_motif_production_counts(
        motif_production_rates_model : jft.Model,
        motif_production_rates_estimate : jft.Vector,
        motif_production_counts : list[jax.Array],
        sample_key : jax.Array,
        minimization_function : Callable
        ):
    """
    Parameters :
    ------------
    motif_production_rate : jax.Array
    motif_production_counts : list[jax.Array]
    sample_key : jax.Array
    minimization_function : Callable
        minimizes nifty8.re.likelihood_impl.Poissonian

    Returns:
    --------
    motif_production_rates_samples : list[jax.Array]
    """
    likelihood = jft.Poissonian(motif_production_counts[0]).amend(motif_production_rates_model)
    for ii in range(1,len(motif_production_counts)):
        likelihood = likelihood + jft.Poissonian(motif_production_counts[ii]).amend(motif_production_rates_model)
    return minimization_function(likelihood, motif_production_rates_estimate)
        

def left_reactant_logconcentration(
        motif_logconcentrations_array : jax.Array,
        motiflength : int,
        number_of_letters : int
    ):
    '''
    if not del_t c_l
    for motiflength = 4:
    collision_exponent[l1,l2,l3,r2,r3,r4][p1,p2,p3,p4] += c[l1,l2,l3,0]
    For jit, motiflength and number_of_letters are stated explicitly,
    the motif_logconcentrations_array must fulfill
    motiflength = len(motif_logconcentrations_array.shape)
    nol = motif_logconcentrations_array.shape[1]

    Parameters:
    -----------
    motif_logconcentrations_array : jax.Array
    motiflength : int
    number_of_letters : int


    Returns:
    --------
    collision_exponent : jax.Array
    '''
    nol = number_of_letters #motif_logconcentrations_array.shape[1]
    collision_exponent = jnp.zeros((nol+1,)*(motiflength-2)+(nol,nol)+(nol+1,)*(motiflength-2)+(nol+1,)*int(motiflength>2)+(nol,nol)+(nol+1,)*(motiflength-3))
    for strandlength in range(1,motiflength):
        shape = (0,)*(motiflength-1-strandlength)+(slice(1,None),)*(strandlength-1)+(slice(None),)
        shape2 = (nol+1,)*(motiflength>1)+(nol,)+(nol+1,)*(motiflength-2)+(1,)*(motiflength-2)#(3*motiflength-strandlength-2)
        shape3 = (slice(1,None),)*int(strandlength>1) + (slice(None),) + (slice(1,None),)*(strandlength-2)+(0,)*(motiflength-strandlength)
        mla = motif_logconcentrations_array.reshape(shape2)
        return mla.at[shape3], shape, collision_exponent
        collision_exponent = collision_exponent.at[shape].add(mla.at[shape3])
    return collision_exponent



import nifty8 as ift
import numpy as np

from ..obj.units import make_unit, Unit

from ..domains.time_space import TimeSpace, are_compatible_timespaces, isinstance_timespace

def TimesVector(times : np.ndarray,
                units : Unit
                ) -> ift.Field:
    return ift.Field(ift.DomainTuple.make(TimeSpace(times, units)),
                     times)

def isinstance_times_vector(obj)->bool:
    if not isinstance(obj,ift.Field):
        print("Not a NIFTy-Field")
        return False
    if not isinstance_timespace(obj.domain):
        print('Domain is not a TimeSpace.')
        return False
    return True

def are_compatible_times_vectors(timesvector1 : TimesVector,
                                 timesvector2 : TimesVector
                                 ) -> bool:
    if not isinstance_times_vector(timesvector1):
        print('Object is not a TimesVector')
        return False
    if not isinstance_times_vector(timesvector2):
        print('Object is not a TimesVector')
        return False
    if not are_compatible_timespaces(timesvector1.domain, timesvector2.domain):
        return False
    return True



import jax.numpy as jnp

from ._rates_utils import _realign_fourmer_concentration_vector
from ._rates_utils import _extract_left_reactant_c, _extract_right_reactant_c, _extract_template_c

def _clip_single_concentration_smoothly(
    concentration,
    soft_reactant_threshold,
    hard_reactant_threshold):
    rtrn = jnp.pi*(concentration-soft_reactant_threshold)/(soft_reactant_threshold-hard_reactant_threshold)
    return (jnp.cos(rtrn) + 1.)/2.

def _clip_concentration_vector_smoothly(
    concentration_vector,
    soft_reactant_threshold,
    hard_reactant_threshold):
    if soft_reactant_threshold==0.:
        return 1.
    weight = jnp.where(
            concentration_vector <= hard_reactant_threshold,
            0.,
            1.)
    return jnp.where(
            weight*(concentration_vector < soft_reactant_threshold),
            _clip_single_concentration_smoothly(concentration_vector,soft_reactant_threshold,hard_reactant_threshold),
            weight).reshape(concentration_vector.shape)

def _transform_rr_to_reactant_format(rr):
    nol = rr.shape[0]-1
    motiflength = (len(rr.shape)+2)/3
    left_reactant_number = right_reactant_number = jnp.sum(nol**jnp.arange(1,motiflength-1))
    template_reactant_number = jnp.sum(nol**jnp.arange(2,motiflength+1))*nol**(motiflength-1)
    # init rr in reactant format (rr_rf)
    rr_rf = jnp.zeros((left_reactant_number,right_reactant_number,template_reactant_number))
    # transform left reactant axes
    for left_strandlength in range(1,motiflength):
        rr_indcs1 = jnp.sum(nol**jnp.arange(1,left_strandlength))
        rr_indcs1 = (slice(rr_indcs1,rr_indcs1+nol**left_strandlength),)
        rr_rf_indcs1 = (0,)*(motiflength-left_strandlength-1) + (slice(None),)*left_strandlength
        for right_strandlength in range(1,motiflength):
            rr_indcs2 = jnp.sum(nol**jnp.arange(1,right_strandlength))
            rr_indcs2 = (slice(rr_indcs2,rr_indcs2+nol**left_strandlength),)
            rr_rf_indcs2 = (slice(None),)*right_strandlength + (0,)*(motiflength-right_strandlength-1)
            for templatelength in  range(2,motiflength+2):
                rr_indcs3 = jnp.sum(nol**jnp.arange(1,templatelength))
                if templatelength == motiflength+1:
                    rr_rf_indcs3 = (slice(1,None),)*strandlength + (0,)
                    rr_indcs3 = (slice(rr_indcs3,rr_indcs3+nol**(templatelength)),)
                else: 
                    rr_rf_indcs3 = (0,) + (slice(1,None),)*templatelength + (0,)*(motiflength-templatelength-1)
                    rr_indcs3 = (slice(rr_indcs3,rr_indcs3+nol**(templatelength+1)),)
                rr_rf_indcs = rr_rf_indcs1 + rr_rf_indcs2 + rr_rf_indcs3
                rr_indcs = rr_indcs1 + rr_indcs2 + rr_indcs3
                rr_rf = rr_rf.at[rr_rf_indcs].set(rr[rr_indcs])
    return rr_rf

def _transform_fourmer_rr_to_reactant_format(rr):
    nol = rr.shape[0]-1
    motiflength = (len(rr.shape)+2)/3
    if motiflength != 4:
        raise NotImplementedError(f"motiflength must be 4, but is {motiflength = }")
    raise NotImplementedError("function not yet implemented.")
    """
    left_reactant_species_number = right_reactant_species_number = nol**jnp.arange(motiflength)
    total_left_reactant_species_number = total_right_reactant_species_number = jnp.sum(left_reactant_species_number)
    template_species_number = jnp.append(nol**jnp.arange(motiflength+1) + nol**(motiflength-1))
    total_template_species_number = jnp.sum(template_species_number)
    rr_rf_shape = (left_reactant_species_number,right_reactant_species_number,template_species_number)
    rr_rf = jnp.zeros(rr_rf_shape)
    # left reactant
    rr_rf[] = rr[0,0,:]
    rr[0,1:,:]
    rr[1:,1:,:]
    # right reactant
    rr[0,1:,0,0]
    rr[0,1:,1:]
    rr[0,1:,1:,1:]
    # template
    rr[0,1:,1:,0]
    rr[0,1:,1:,1:]
    rr[1:,1:,1:,1:]
    rr[1:,1:,1:,0]
    """

def _clip_smoothly(
        reactant_concentration,
        soft_reactant_threshold : float,
        hard_reactant_threshold : float = None
        ) -> jnp.ndarray:
    if soft_reactant_threshold == 0.:
        return 1.
    if hard_reactant_threshold is None:
        hard_reactant_threshold = soft_reactant_threshold/2.
    """
    weight = jnp.where(
            reactant_concentration <= hard_reactant_threshold,
            0.,
            1.)
    weight = jnp.where(
            weight*(reactant_concentration < soft_reactant_threshold),
            _clip_single_concentration_smoothly(reactant_concentration,soft_reactant_threshold,hard_reactant_threshold),
            weight)
    """
    left_reactant_c = _extract_left_reactant_c(
            _realign_fourmer_concentration_vector(reactant_concentration)
            )
    right_reactant_c = _extract_right_reactant_c(reactant_concentration)
    template_c = _extract_template_c(reactant_concentration)
    # left weight
    left_weight = _clip_concentration_vector_smoothly(left_reactant_c, soft_reactant_threshold, hard_reactant_threshold)
    # right_weight
    right_weight = _clip_concentration_vector_smoothly( right_reactant_c, soft_reactant_threshold,hard_reactant_threshold)
    # template_weight
    template_weight = _clip_concentration_vector_smoothly(template_c, soft_reactant_threshold,hard_reactant_threshold)
    return jnp.outer(
            jnp.outer(left_weight,right_weight),
            template_weight
            ).reshape(
                    left_weight.shape+right_weight.shape+template_weight.shape
                    )



from .motif_vector import MotifVector
from .units import make_unit

def MotifConcentrationVector(motiflength : int,
                      alphabet : str):
    unit = make_unit('mol')/make_unit('L')
    return MotifVector(motiflength, alphabet, unit)



from collections import namedtuple
import numpy as  np
import jax.numpy as jnp
import jax
import yaml
from typing import Union

from warnings import warn

import nifty8 as ift

from .motif_vector import _motif_vector_as_array, _array_to_motif_vector_dct, category_indices
from .motif_vector import MotifVector, isinstance_motifvector

from ..domains.trajectory_space import TrajectorySpace, TimeRGSpace, UnstructuredTimeDomain
from .times_vector import TimesVector
from .units import isunit, make_unit
from .units import transform_unit_to_dict, transform_dict_to_unit

from ..utils.save import create_directory_path_if_not_already_existing
from ..utils.transform_vectors_to_fields import transform_dicts_to_field

def MotifTrajectory(motif_vectors : list,
                    times : Union[TimeRGSpace, UnstructuredTimeDomain]):
    """
    """
    motif_vector = motif_vectors[0]
    motifs = [motif_vectors[ii].motifs for ii in range(len(motif_vectors))]
    if not isinstance(times,ift.Field):
        raise TypeError('times vector needs to be Field')
    if not (isinstance(times.domain,(TimeRGSpace,UnstructuredTimeDomain)) or isinstance(times.domain[0],(TimeRGSpace,UnstructuredTimeDomain))):
        raise TypeError('times vector domain needs to be a time domain')
    motif_trajectory_properties = {'motiflength' : motif_vector.motiflength,
        'alphabet' : motif_vector.alphabet,
        'number_of_letters' : motif_vector.number_of_letters,
        'unit' : motif_vector.unit,
        'times': times,
        'motifs' : transform_dicts_to_field(motifs, times.domain)}
    for motif_vector in motif_vectors:
        assert(motif_vector.motiflength==motif_trajectory_properties['motiflength'])
        assert(motif_vector.alphabet==motif_trajectory_properties['alphabet'])
        assert(motif_vector.number_of_letters==motif_trajectory_properties['number_of_letters'])
        if isunit(motif_vector.unit)==isunit(motif_trajectory_properties['unit']):
            assert motif_vector.unit==motif_trajectory_properties['unit'], "Units of motif vectors need to be the same:{unit1} != {unit2}".format(
                unit1 = motif_vector.unit,
                unit2 = motif_trajectory_properties['unit']
                )
        else:
            raise AssertionError("Units of motif vectors need to be the same:{unit1} != {unit2}".format(
                unit1 = motif_vector.unit,
                unit2 = motif_trajectory_properties['unit']
                )
                                 )
    mt = namedtuple('MotifTrajectory',
                tuple(motif_trajectory_properties.keys()))
    return mt(**motif_trajectory_properties)

def isinstance_motiftrajectory(obj) -> bool:
    is_motif_trajectory = True
    keys = ['motiflength', 'alphabet', 'number_of_letters', 'unit', 'times', 'motifs']
    for key in obj._asdict().keys():
        if key not in keys:
            print('Not a MotifTrajectory, missing key: {}.'.format(key))
            return False
    is_motif_trajectory *= isinstance(obj, tuple)
    is_motif_trajectory *= hasattr(obj, '_asdict')
    is_motif_trajectory *= hasattr(obj, '_fields')
    return bool(is_motif_trajectory)

def are_compatible_motif_trajectories(trj1 : MotifTrajectory, trj2 : MotifTrajectory) -> bool:
    if not isinstance_motiftrajectory(trj1):
        print('Object is not a MotifTrajectory')
        return False
    if not isinstance_motiftrajectory(trj2):
        print('Object is not a MotifTrajectory')
        return False
    keys = ['motiflength', 'alphabet', 'unit']
    for key in keys:
        if not np.prod(trj1._asdict()[key]==trj2._asdict()[key]):
            print('MotifTrajectories not compatible: {} mismatch.'.format(key))
            return False
    try:
        if not np.prod(trj1.times.val==trj2.times.val) or not np.prod(trj1.times.domain[0]==trj2.times.domain[0]):
            key = 'times'
            warn('MotifTrajectories {} mismatch.'.format(key))
    except ValueError:
        key = 'times'
        warn('MotifTrajectories {} mismatch.'.format(key))
    return True

def save_motif_trajectory(archive_path : str,
        motif_trajectory : MotifTrajectory
        ) -> None:
    create_directory_path_if_not_already_existing(archive_path)
    motifs, times = _motif_trajectory_as_array(motif_trajectory)
    jnp.save(archive_path+'motif_trajectory_motifs',
            motifs
            )
    jnp.save(archive_path+'motif_trajectory_times',
            times
            )

    with open(archive_path+'motif_trajectory_properties.yaml','w') as yaml_file:
        yaml.dump({'motiflength':motif_trajectory.motiflength,
            'alphabet':motif_trajectory.alphabet,
                   'unit':transform_unit_to_dict(motif_trajectory.unit),
                   'times_unit' : transform_unit_to_dict(motif_trajectory.times.domain[0].units)},
            yaml_file,
            indent=4)

def load_motif_trajectory(archive_path : str
        ) -> MotifTrajectory:
    dct_filename = archive_path+'motif_trajectory_properties.yaml'
    motifs_filename = archive_path+'motif_trajectory_motifs'+'.npy'
    times_filename = archive_path+'motif_trajectory_times.npy'
    with open(dct_filename, 'r') as yaml_file:
        motif_trajectory_properties = yaml.safe_load(yaml_file)
    times_unit = motif_trajectory_properties['times_unit']
    del motif_trajectory_properties['times_unit']
    motif_trajectory_properties['unit'] = transform_dict_to_unit(motif_trajectory_properties['unit'])
    makeMotifVector = MotifVector(**motif_trajectory_properties)
    motif_vector_list = [makeMotifVector(
        _array_to_motif_vector_dct(motif_vec,motif_trajectory_properties['motiflength'],motif_trajectory_properties['alphabet'])
        ) for motif_vec in jnp.load(motifs_filename)]
    times_vector = TimesVector(jnp.load(times_filename),times_unit)
    return MotifTrajectory(
            motif_vector_list,
            times=times_vector
            )

def _array_to_motif_trajectory(
        motif_trajectory_array : np.ndarray,
        times : TimesVector,
        **motif_trajectory_properties
        ) -> MotifTrajectory:
    motif_vectors = [MotifVector(
            motif_trajectory_properties['motiflength'],
            motif_trajectory_properties['alphabet'],
            motif_trajectory_properties['unit']
            )(_array_to_motif_vector_dct(
                motif_trajectory_array[ii],
                motif_trajectory_properties['motiflength'],
                motif_trajectory_properties['alphabet'],
                ))
                     for ii in range(len(motif_trajectory_array))] 
    return MotifTrajectory(motif_vectors, times)

def _motif_trajectory_as_array(motif_trajectory : MotifTrajectory
        ) -> Union[np.ndarray,np.ndarray]:
    """
    returns motifs and times of MotifTrajectory object into numpy arrays.

    Parameters:
    -----------
    motif_trajectory : MotifTrajectory

    Returns:
    --------
    motifs : np.ndarray
    times : np.ndarray
    """
    times = np.asarray(motif_trajectory.times.val)
    number_of_letters = motif_trajectory.number_of_letters
    motiflength = motif_trajectory.motiflength

    motif_array_shape = (number_of_letters+1,)
    motif_array_shape += (number_of_letters,)*int(motiflength>2)
    motif_array_shape += (number_of_letters+1,)*(motiflength-1-int(motiflength>2))
    motif_array = np.zeros((len(times),) + motif_array_shape)
    for motif_category in motif_trajectory.motifs.keys():
        # get indices
        indices = (slice(None),) + category_indices(motif_category, motiflength, motif_trajectory.alphabet)
        motif_array[indices] += motif_trajectory.motifs[motif_category].val
    return motif_array, times

def extract_initial_motif_vector_from_motif_trajectory(
        motif_trajectory : MotifTrajectory,
        c_ref : float = None
    ) -> MotifVector:
    motif_array, _ = _motif_trajectory_as_array(motif_trajectory)
    if c_ref is not None:
        n_ref = motif_array[0,0,0,0,0]
        motif_array = c_ref*motif_array/n_ref
        unit = make_unit('c_0')
    else:
        unit = motif_trajectory.unit
    mvd = _array_to_motif_vector_dct(
        motif_array[0],
        motif_trajectory.motiflength,
        motif_trajectory.alphabet
    )
    motif_vector = MotifVector(
        motif_trajectory.motiflength,
        motif_trajectory.alphabet,
        unit
    )
    return motif_vector(mvd)



import morsaik as kdi
import numpy as np

def test_motif_breakage_trajectory():
    motiflengths = [2,3,4,5,6,9,10]
    alphabet = ['a','b']
    units = [kdi.make_unit(''),1./kdi.make_unit('s')]
    times = kdi.TimesVector(np.arange(3), kdi.make_unit('s'))
    for motiflength in motiflengths:
        for unit in units:
            motif_breakage_vector_dct = kdi._create_empty_motif_breakage_dct(motiflength,
                                                alphabet
                                                )
            motif_breakage_vector = kdi.MotifBreakageVector(motiflength,alphabet,unit)
            motif_breakage_vector = motif_breakage_vector(motif_breakage_vector_dct)
            mbvs = [motif_breakage_vector]*times.shape[0]
            motif_breakage_trajectory = kdi.MotifBreakageTrajectory(mbvs, times)
            assert(kdi.isinstance_motifbreakagetrajectory(motif_breakage_trajectory))
            assert(kdi.are_compatible_motif_breakage_trajectories(
                motif_breakage_trajectory,
                motif_breakage_trajectory)
                   )



import nifty8 as ift
import numpy as np

from ..obj.motif_vector import isinstance_motifvector
from ..obj.motif_production_vector import isinstance_motifproductionvector

from ..domains.trajectory_space import TrajectorySpace

def transform_dicts_to_field(motifs : list,
                             time_space : ift.DomainTuple
                             ) -> ift.MultiField:
    trajectory_domain = TrajectorySpace(motifs[0].domain, time_space)
    field = {}
    for key in motifs[0].keys():
        #np.zeros(time_space.shape+motif_list[0].motifs[key].shape)
        field[key] = np.asarray([motifs[ii][key].val for ii in range(len(motifs))])
        field[key] = ift.Field(
            trajectory_domain[key],
            field[key]
            )
    return ift.MultiField.from_dict(field, domain=trajectory_domain)

def transform_vectors_to_field(motif_list : list,
                                  time_space : ift.DomainTuple) -> ift.MultiField:
    if isinstance_motifvector(motif_list[0], print_statements = False):
        motifs = [motif_list[ii].motifs for ii in range(len(motif_list))]
    elif isinstance_motifproductionvector(motif_list[0]):
        motifs = [motif_list[ii].productions for ii in range(len(motif_list))]
    #elif isinstance_motifbreakagevector(motif_list[0]):
    #    motifs = [motif_list[ii].breakages for ii in range(len(motif_list))]
    else:
        raise TypeError("motif_list needs to be MotifTrajectory or MotifProductionTrajectory")
    return transform_dicts_to_field(motifs, time_space)



def split_config(config : dict, split_keys : list):
    """
    splits config dictionary infto list of config files
    by checking how many parameter sets there are for the split_keys
    and separating the config files at those split_keys.
    The keys of the config dictionary that are not split_keys,
    are the same for every config in the resulting list.
    """
    from copy import deepcopy
    splitted_config_len = len(config[split_keys[0]])
    splitted_config = [deepcopy(config) for jj in range(splitted_config_len)]

    key_values = {}
    for split_key in split_keys:
        vl = deepcopy(config[split_key])
        if isinstance(vl, list) and len(vl)==splitted_config_len:
            key_values[split_key] = vl
        else:
            key_values[split_key] = [vl,]*splitted_config_len

    for jj in range(splitted_config_len):
        for split_key in split_keys:
            splitted_config[jj][split_key] = key_values[split_key][jj]

    return splitted_config



import jax
import jax.numpy as jnp
import numpy as np

from .fourmer_production_rates import compute_motif_extensions, _set_invalid_logc_to_log0

def collisions_from_motif_concentration_trajectory_and_collision_rate_constants(
    motif_concentration_trajectory,
    collision_rate_constants = 1,
    flatten : bool = True
    ):
    '''
    The collisions give how often a certain ligation (channel)
    can be observed.
    It measures, how often the reactants (end motif of the rear strand,
    beginning motif of the fore strand and template motif)
    that perform the reaction, meet (are exposed to each other).

    Technically, it integrates over time the three
    concentrations reaction wise,
    according to effective ligation dynamics.

    Parameters:
    -----------
    motif_concentration_trajectory  :   concentration
    collision_rate_constants:   array, optional (default: None)
        weights of the integral. If None, they are set to one.
    flatten :   boolean, optional (default: True)
        if True, output is flat

    Returns:
    --------
    field : NIFTy-field
        field with each entry being the integral over the corresponding
        concentrations of the reactants that are needed for the templated
        ligation.
    '''
    #'length{}strand','beginning','continuation','end'
    endings_trajectory = motif_concentration_trajectory.motifs['ending']
    beginnings_trajectory = motif_concentration_trajectory.motifs['beginning']
    templates_trajectory = motif_concentration_trajectory.motifs

    return collisions

def motifs_collisions_array_from_motifs_array(
        motifs_array : jax.Array,
        collision_order : int = 3
        ) -> jax.Array:
    """
    takes motifs_array and calculates motifs_collisions_array, i.e.
    the $m$th outer product of the motifs array with itself,
    where $m$ is the collision_order.

    Parameters
    ----------
    motifs_array : jax.Array
    collision_order : int, optional
        number of collisions, by default 3

    Returns
    -------
    jax.Array
        motifs_collisions_array
    """
    motifs_collisions_array = motifs_array
    for ii in range(1,collision_order):
        motifs_collisions_array = motifs_collisions_array.reshape(-1,1)*motifs_array.reshape(-1,1).T
    return motifs_collisions_array.flatten()

def ligation_spot_formations_from_motifs_array(
        motifs_array : jax.Array,
        number_of_letters : int = 4,
        motiflength : int = 4
        ) -> jax.Array:
    """
    calculates the collisions that actually enable ligation,
    i.e. the $m$th outer product of the motifs array with itself,
    using infer.motifs_collisions_array_from_motifs_array
    but with only ending motifs (and strands) as fore motifs,
    beginning motifs (and strands) as rear motifs
    and only strands of at least lenth 2 as template motifs.

    Parameters
    ----------
    motifs_array : jax.Array
    number_of_letters : int, optional
        number of letters, by default 4
    motiflength : int, optional
        length of motif, by default 4

    Returns
    -------
    jax.Array
        ligation_spot_array
    """
    ligation_spot_array = motifs_collisions_array_from_motifs_array(motifs_array, collision_order = 3)
    number_of_motifs = jnp.sum(number_of_letters**jnp.arange(1,motiflength+1))+number_of_letters**(motiflength-1)
    ligation_spot_array = ligation_spot_array.reshape((number_of_motifs,)*3)
    # not all collisions enable adjacent hybridization
    number_of_continuations = number_of_letters**motiflength
    number_of_beginnings = number_of_endings = number_of_letters**(motiflength-1)
    # fore motif must end
    ligation_spot_array = jnp.concatenate(
            [ligation_spot_array[:-(number_of_beginnings + number_of_endings + number_of_continuations),:,:],
             ligation_spot_array[-number_of_endings:,:,:]
             ],
            axis = 0)
    # rear motif must begin
    ligation_spot_array = ligation_spot_array[:,:-(number_of_continuations+number_of_endings),:]
    # template must have at least one covalent bond
    return ligation_spot_array[:,:,number_of_letters:]

def collisions_from_motif_concentration_trajectory_array_and_collision_rate_constants_array(
    motif_concentration_trajectory_array : jax.Array,
    motif_concentration_trajectory_times_array : jax.Array = None,
    collision_rate_constants_array : jax.Array = 1.,
    motiflength : int = 4,
    alphabet : list = ['A', 'T'],
    complements : list = [1,0],
    concentrations_are_logarithmised : bool = True,
    pseudo_count_concentration : float = 1.e-12
    ):
    """
    Jax version of collions_from_motif_concentration_trajectory_and_collision_rate_constants
    """
    if motiflength != 4:
        raise NotImplementedError("Function only implemented for motiflength 4.")
    nol = len(alphabet)
    # for arbitrary motiflength:
    # left_reactants = motif_concentration_trajectory_array.at[(slice(None),)+(slice(None),)*(motiflength-1)+(0,)]
    # right_reactants = motif_concentration_trajectory_array.at[(slice(None),)+(0,)]
    # templates = motif_concentration_trajectory_array.at[(slice(None),)+(slice(None),)*(motiflength//2-2)+(slice(None),slice(1,None))]
    # collisions = jnp.zeros((nol+1,)*(motiflength-2)+(nol,nol)+(nol+1,)*(motiflength-2)+(nol+1,)+(nol))
    # for strandlength in range(1,motiflength-1):
    # ...
    logc = jnp.zeros((len(jnp.asarray(motif_concentration_trajectory_times_array)),)+(nol+1,)*motiflength)
    if not concentrations_are_logarithmised:
        motif_concentration_trajectory_array = motif_concentration_trajectory_array.at[motif_concentration_trajectory_array<pseudo_count_concentration].set(pseudo_count_concentration)
        logc = logc.at[:,:,1:,:,:].set(jnp.log(motif_concentration_trajectory_array))
        concentrations_are_logarithmised = True
    else:
        logc = logc.at[:,:,1:,:,:].set(motif_concentration_trajectory_array)
    if not isinstance(collision_rate_constants_array,(jax.Array,list)):
        collision_rate_constants_array = collision_rate_constants_array*jnp.ones((nol+1,)*(2*motiflength))

    collisions = jnp.array([[
        compute_motif_extensions(
            jnp.asarray(logc)[ii],#logc or c
            basic_rate_constants = jnp.asarray(collision_rate_constants_array),
            motiflength = motiflength,
            number_of_letters = len(alphabet)
            )[:,1:,1:,:,:,1:,1:,:].reshape(-1)
         ] for ii in range(len(motif_concentration_trajectory_times_array))])

    collisions = collisions.reshape(collisions.shape[0],-1)
    collisions = jnp.trapezoid(
        collisions,
        x=motif_concentration_trajectory_times_array,
        axis=0
        )
    # collisions_= collisions.reshape()
    return collisions



import numpy as np
from nifty8 import DomainTuple
from nifty8.utilities import check_object_identity, frozendict, indent
import itertools
from typing import Union
from ..obj.units import make_unit

from .hamming_space import HammingSpace
from .motif_space import MotifSpace
from .motif_space import _return_motif_categories
# MotifDomain -> MotifSpace
# ExtensionRates, Motif fields
# st = StrandTrajectory
## -> motif concentrations(time) DomainTuple((TimeSpace,MotifSpace),
##    motif productions(time): DomainTuple((TimeSpace,MPRCSpace)
##    length distribution
##    parameters
##    
# st.motif_trajectory, st.ligation_counts, st.params = st(complexes, ligation_statistics, parameters)
# te = TrajectoryEnsemble([st for st in sts])
# ri = RateInference(method,...)
# extension_rates = ri(te)
# er = extension_rates 
# oi = ODEIntegrator(...)
# mt = oi(er,initial_motifs)

def _return_motif_production_categories(motiflength : int,
        alphabet : list,
        maximum_ligation_window_length : int
        ) -> tuple:
    return tuple(make_motif_production_dct(
        alphabet,
        motiflength,
        maximum_ligation_window_length).keys()
        )

def _production_channel_id(product_category : str,
        template_category : str,
        ligation_window_length : int,
        ligation_spot : int) -> str:
    """
    The production channel id is given by the categories of the product
    and the template as well as the ligation window length and the ligation
    spot.
    The ligation window is set such that it captures the whole hybridization
    site up to one dangling nucleotide, this means, all spots are occupied by
    letters, but the very outer ones which can (but don't have to) also be
    empty.
    """
    return product_category + '_{}_'.format(ligation_window_length) + '{}_'.format(ligation_spot) + template_category

def _determine_product_and_template_categories_and_ligation_spots(motiflength : int,
        maximum_ligation_window_length : int,
        ligation_window_length : int
        ) -> Union[list,list]:
    if maximum_ligation_window_length < 4:
        ligation_spots = np.array([maximum_ligation_window_length])-2

        product_categories = [product_category for product_category in _return_motif_categories()[-2-int(motiflength>2):-1]]
        template_categories = [template_category for template_category in _return_motif_categories()[-2:-1 if (motiflength<3) else None]]
    else:
        ligation_spots = np.arange(1,ligation_window_length-2)

        product_categories = [product_category.format(ligation_window_length-2) for product_category in _return_motif_categories()]
        template_categories = product_categories
    return product_categories, template_categories, ligation_spots

def _valid_production_channel(product_category : str,
        template_category : str,
        ligation_window_length : int,
        ligation_spot : int,
        maximum_ligation_window_length : int
        ) -> bool:
    if ligation_window_length < maximum_ligation_window_length:
        if (product_category, template_category) in itertools.product(_return_motif_categories()[-3:-1],_return_motif_categories()[-2:]):
            return False
        if (product_category, template_category) in itertools.product(_return_motif_categories()[-2:],_return_motif_categories()[-3:-1]):
            return False
    elif (product_category in _return_motif_categories()[-2:]
            and template_category in _return_motif_categories()[-3:-1]
            and ligation_spot<ligation_window_length-ligation_window_length//2-1
            ):#FIXME
        return False
    elif (product_category in _return_motif_categories()[-3:-1]
            and template_category in _return_motif_categories()[-2:]
            and ligation_spot>ligation_window_length-ligation_window_length//2-1
            ):#FIXME
        return False
    return True

def make_motif_production_dct(alphabet : list,
        motiflength : int,
        maximum_ligation_window_length : int,
        ) -> dict:
    """
    Creates a motif production dictionary.
    For the production channel id, see morsaik.domains._production_channel_id
    The main concept behind the productions is that the hybridization
    and the ligation kinetics are only characterized by the hybridization sites
    and maximally one further nucleotide of a dangling strand.
    Thus, the motif productions are specified by a ligation window that
    captures the exact hybridization and this potentially dangling end.
    In the keys, we iterate over all possible ligation windows with different
    length.
    The maximum length can be set optionally,
    else it is the motiflength, which is also the maximum
    maximum_ligation_window_length, one can set.
    For the maximum_ligation_window_length, also fully occupied strands are
    tracked, for ligation_window_lengths that are smaller than the
    maximum_ligation_window_length, those are not tracked, since they are
    considered by larger ligation windows.
    """
    if maximum_ligation_window_length is None:
        maximum_ligation_window_length = motiflength
    elif maximum_ligation_window_length > motiflength:
        raise ValueError("Maximum ligation window length needs to be smaller or equal to the motif length")
    if maximum_ligation_window_length < 4:
        ligation_window_lengths = np.array([maximum_ligation_window_length])
    else:
        ligation_window_lengths = np.arange(4,maximum_ligation_window_length+1)
    domain_dct = {}
    for ligation_window_length in ligation_window_lengths:
        product_categories, template_categories, ligation_spots = _determine_product_and_template_categories_and_ligation_spots(motiflength,
                maximum_ligation_window_length,
                ligation_window_length
                )

        central_produced_motif_space = MotifSpace.make(alphabet,ligation_window_length,monomers_included=False)
        template_motif_space = MotifSpace.make(alphabet,ligation_window_length,monomers_included=False)

        for product_category, template_category, ligation_spot in itertools.product(product_categories, template_categories, ligation_spots):
            if not _valid_production_channel(product_category, template_category,
                ligation_window_length, ligation_spot,
                maximum_ligation_window_length):
                continue
            product_space = central_produced_motif_space[product_category]
            template_space = template_motif_space[template_category]
            reaction_key = _production_channel_id(product_category, template_category,
                    ligation_window_length, ligation_spot)
            domain_dct[reaction_key] = DomainTuple.make(product_space[:]+template_space[:])
    return domain_dct

class MotifProductionSpace(MotifSpace):
    """
    Domain of Motif Productions,
    like motif production rate constants,
    ligations projected onto motif space.
    """
    def __init__(self, dct,
            alphabet = None,
            motiflength = None,
            maximum_ligation_window_length=None,
            units ='bits',
            _callingfrommake=False):
        if not _callingfrommake:
            raise NotImplementedError(
                'To create a MultiDomain call `MultiDomain.make()`.')
        self._alphabet = alphabet
        self._motiflength = motiflength
        self._units = make_unit(units)
        self._maximum_ligation_window_length = motiflength if maximum_ligation_window_length is None else maximum_ligation_window_length
        self._keys = tuple(sorted(dct.keys()))
        self._domains = tuple(dct[key] for key in self._keys)
        self._idx = frozendict({key: i for i, key in enumerate(self._keys)})

    @staticmethod
    def make(alphabet : list,
             motiflength : int,
             maximum_ligation_window_length : int
             ) -> object:
        """Creates a MultiDomain object from a dictionary of names and domains

        Parameters
        ----------
        inp : MultiDomain or dict{name: DomainTuple}
            The already built MultiDomain or a dictionary of DomainTuples

        Returns
        ------
        A MultiDomain with the input Domains as domains
        """
        if isinstance(alphabet,int):
            alphabet = list(1 + np.arange(alphabet))

        inp = make_motif_production_dct(alphabet,
                motiflength,
                maximum_ligation_window_length)

        if isinstance(inp, MotifProductionSpace):
            return inp
        if not isinstance(inp, dict):
            raise TypeError("dict expected")
        tmp = {}
        for key, value in inp.items():
            if not isinstance(key, str):
                raise TypeError("keys must be strings")
            tmp[key] = DomainTuple.make(value)
        tmp = frozendict(tmp)
        obj = MotifProductionSpace._domainCache.get(tmp)
        if obj is not None:
            return obj
        obj = MotifProductionSpace(tmp, alphabet=alphabet, motiflength=motiflength, maximum_ligation_window_length=maximum_ligation_window_length, _callingfrommake=True)
        MotifProductionSpace._domainCache[tmp] = obj
        return obj

    @property
    def alphabet(self):
        return self._alphabet

    @property
    def number_of_letters(self):
        return len(self._alphabet)

    @property
    def maximum_ligation_window_length(self):
        return self._maximum_ligation_window_length

    @property
    def motiflength(self):
        return self._motiflength



from os.path import exists

from .alphabet import alphabet as get_alphabet
from ..read.strand_motifs_productions_trajectory_ensemble import strand_motifs_productions_trajectory_ensemble as read_strand_motifs_productions_trajectory_ensemble
from ..utils.manage_strand_reactor_files import _create_typical_strand_trajectory_section_ligations_filepath 

def strand_motifs_productions_trajectory_ensemble(
        motiflength : int,
        strand_trajectory_id : str,
        param_file_no : int = 0,
        skiprows : int = 2,
        maximum_ligation_window_length : int = None,
        ):
    if maximum_ligation_window_length is None:
        maximum_ligation_window_length = motiflength

    alphabet = get_alphabet(strand_trajectory_id)

    current_filepath = lambda srn, sn : _create_typical_strand_trajectory_section_ligations_filepath(
            strand_trajectory_id,
            param_file_no=param_file_no,
            simulations_run_no=srn,
            simulations_no = sn,
            )

    simulations_run_no = 0
    simulations_no = 0
    filepath_lists = []
    while exists(current_filepath(simulations_run_no, simulations_no)):
        filepaths = []
        while exists(current_filepath(simulations_run_no, simulations_no)):
            filepaths += [current_filepath(simulations_run_no, simulations_no),]
            simulations_no += 1
        filepath_lists = filepath_lists + [filepaths,]
        simulations_run_no += 1
    return read_strand_motifs_productions_trajectory_ensemble(filepath_lists,
            alphabet,
            motiflength,
            maximum_ligation_window_length,
            skiprows =skiprows)



import numpy as np
import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap

def _choose_color(
        color : LinearSegmentedColormap,
        color_parameter : float = 0
        ):
    if isinstance(color,LinearSegmentedColormap):
        return color(color_parameter)
    else:
        return color

def standard_colorbar():
    green = mcolors.CSS4_COLORS['darkgreen']
    darkgold = mcolors.CSS4_COLORS['darkgoldenrod']
    gold = mcolors.CSS4_COLORS['gold']
    edgecolors = mcolors.to_rgba_array([green,darkgold,gold])[:,:-1]
    return LinearSegmentedColormap.from_list('greengold',edgecolors,100)

def greenish_colorbar():
    darkgreen = mcolors.CSS4_COLORS['darkslategray']
    forestgreen = mcolors.CSS4_COLORS['forestgreen']
    lightgreen = mcolors.CSS4_COLORS['yellow']
    edgecolors = mcolors.to_rgba_array([darkgreen,forestgreen,lightgreen])[:,:-1]
    return LinearSegmentedColormap.from_list('darklightgreenish',edgecolors,100)

def green_colorbar():
    darkgreen = mcolors.CSS4_COLORS['darkgreen']
    forestgreen = mcolors.CSS4_COLORS['forestgreen']
    lightgreen = mcolors.CSS4_COLORS['lightgreen']
    edgecolors = mcolors.to_rgba_array([darkgreen,forestgreen,lightgreen])[:,:-1]
    return LinearSegmentedColormap.from_list('darklightgreen',edgecolors,100)

def colorbar_example():
    greengold_cmapp = standard_colorbar()
    data = np.linspace(1.,0.,1048)
    data = np.outer(data,np.ones(data.shape))
    fig,axs = plt.subplots(1,1)
    im = axs.imshow(data,cmap=greengold_cmapp)
    fig.colorbar(im,ax=axs)



from typing import Union
import numpy as np
import warnings
import itertools

"""
For a segment, the hybridization energy is computed by 2x2 blocks.
The hybridization energy of a 2x2 block is given by one of the following
hybridization energy parameters:
    0 dG_4_2Match_mean
    1 dG_4_1Match
    2 dG_4_0Match
    3 dG_3_1Match_mean
    4 dG_3_0Match
    5 ddG_4_2Match_alternating
    6 ddG_3_1Match_alternating
They are stored in the hybridization_energy_parameters vector
"""
def segemnt_hybridization_energy(
        segment,
        hybridization_energy_parameters
        ):
    segment_shape = (len(alphabet),)*(2*motiflength//2)


import jax
import jax.numpy as jnp

def build_coefficient_tensor_lookup(f_lookup, n):
    """
    f_lookup: integer array of shape (n+1,) giving f(i).
    Output: coefficient array C[ii,jj,kk,ll,k] with shape (n+1,n+1,n+1,n+1,7)
    """
    f_lookup = jnp.asarray(f_lookup)

    idx = jnp.arange(n + 1)
    ii, jj, kk, ll = jnp.meshgrid(idx, idx, idx, idx, indexing="ij")

    ii0 = (ii == 0)
    jj0 = (jj == 0)
    kk0 = (kk == 0)
    ll0 = (ll == 0)
    num_zero = ii0 + jj0 + kk0 + ll0

    # vectorized f
    f_ii = f_lookup[ii]
    f_kk = f_lookup[kk]

    cond_fii_eq_jj = (f_ii == jj)
    cond_fkk_eq_ll = (f_kk == ll)

    C = jnp.zeros((n+1, n+1, n+1, n+1, 7), dtype=jnp.float32)

    # CASE 1 (exactly one zero)
    case1 = (num_zero == 1)

    # g31 branch (k=3) vs g30 branch (k=4)
    g31_branch = (cond_fii_eq_jj | cond_fkk_eq_ll) & case1
    g30_branch = (~(cond_fii_eq_jj | cond_fkk_eq_ll)) & case1

    C = C.at[...,3].set(g31_branch.astype(jnp.float32))
    C = C.at[...,4].set(g30_branch.astype(jnp.float32))

    # g31a correction (k=6)
    pos_cond = (
        ((ii != kk) & (~ii0) & (~kk0)) |
        ((jj != ll) & (~jj0) & (~ll0))
    ) & g31_branch

    neg_cond = (
        ((ii == kk) & (~ii0)) |
        ((jj == ll) & (~jj0))
    ) & g31_branch

    C = C.at[...,6].set(
        jnp.where(pos_cond,  0.5,
        jnp.where(neg_cond, -0.5, 0.0))
    )

    # CASE 2 (no zeros)
    case2 = (num_zero == 0)

    g40_case = (~(cond_fii_eq_jj | cond_fkk_eq_ll)) & case2
    g41_case = (cond_fii_eq_jj ^ cond_fkk_eq_ll) & case2
    g42_case = (cond_fii_eq_jj & cond_fkk_eq_ll) & case2

    C = C.at[...,2].set(g40_case.astype(jnp.float32))
    C = C.at[...,1].set(g41_case.astype(jnp.float32))
    C = C.at[...,0].set(g42_case.astype(jnp.float32))

    # g42a correction (k=5)
    g42a_corr = jnp.where(ii == kk, -0.5, 0.5)
    C = C.at[...,5].set(jnp.where(g42_case, g42a_corr, 0.0))

    return C



from typing import Union
import numpy as np
import warnings
import itertools

def hybridization_bond_category(
        upper_nucleotide : int,
        hybridized_nucleotide : int,
        complements : list
    ) -> (int, bool):
    number_of_mismatches = 0
    hybridization_is_dangling_end = False
    complements_0 = [0,] + [complement+1 for complement in complements]
    if upper_nucleotide * hybridized_nucleotide == 0: # if any empty letter:
        if upper_nucleotide + hybridized_nucleotide == 0: # if all empty letters:
            raise ValueError("Empty Block")
        else:
            hybridization_is_dangling_end = True
    elif upper_nucleotide != complements_0[hybridized_nucleotide]:
        number_of_mismatches += 1
    return number_of_mismatches, hybridization_is_dangling_end

def segment_category(upper_nucleotide1 : int,
                     hybridized_nucleotide1 : int,
                     upper_nucleotide2 : int,
                     hybridized_nucleotide2 : int,
                     complements : list
                     ) -> (int, bool, int, bool):
    left_mismatch, dangling_end1 = hybridization_bond_category(upper_nucleotide1, hybridized_nucleotide1, complements)
    right_mismatch, dangling_end2 = hybridization_bond_category(upper_nucleotide2, hybridized_nucleotide2, complements)
    return left_mismatch, right_mismatch, dangling_end1, dangling_end2

def analyze_segment(ii,jj,kk,ll,complements):
    number_of_mismatches, number_of_mismatches2, dangling_end1, dangling_end2 = segment_category(ii,jj,kk,ll, complements)
    number_of_mismatches += number_of_mismatches2
    return number_of_mismatches, dangling_end1, dangling_end2

def dissociation_constant_from_strand_reactor_parameters(
        #strand_reactor_parameters : dict,
        characteristic_ligation_length : int,
        ligation_parameters : dict,
        complements : list,
        hybridization_parameters : dict,
        standard_concentration : float = 1.,
        temperature = 1.#k_B K,
    ) -> np.ndarray:
    number_of_letters = len(complements)
    motif_shape = (number_of_letters+1,number_of_letters)
    motif_shape += motif_shape[::-1]
    total_free_energy = _total_free_energy(motif_shape, complements, hybridization_parameters, temperature, number_of_letters)
    # Nochmal checken mit der Standard-Konzentration
    return standard_concentration**2*np.exp(total_free_energy/temperature)

def ligation_rate_constant_from_strand_reactor_parameters(
        hybridization_parameters : dict,
        ligation_parameters : dict,
        complements : list,
        characteristic_ligation_length : int
    ):
    number_of_letters = len(complements)
    motif_shape = give_shape(number_of_letters)
    average_energy_value_of_complementary_blocks = hybridization_parameters['dG_4_2Match_mean']
    stalling_factor = _calculate_stalling(motif_shape, number_of_letters, ligation_parameters, complements)
    return np.exp(average_energy_value_of_complementary_blocks * characteristic_ligation_length) * stalling_factor 

def template_averaged_dissociation_constant_from_strand_reactor_parameters(
        characteristic_ligation_length,
        ligation_parameters,
        complements,
        hybridization_parameters,
        standard_concentration,
        temperature = 1.#k_B K,
    ):
    discon = dissociation_constant_from_strand_reactor_parameters(
        characteristic_ligation_length,
        ligation_parameters,
        complements,
        hybridization_parameters,
        standard_concentration,
        temperature
    )
    return 1./np.sum(1./discon, axis=(-4,-3,-2,-1))


def energy_continuous_block(ii,jj,kk,ll, complements, parameters):
    # check that block is actually a block and not a blunt end
    if ii+jj==0 or kk+ll==0:
        return 0.
    number_of_mismatches, dangling_end, dangling_end2 = analyze_segment(ii,jj,kk,ll, complements)
    # check that block is actually continuous, not dangling on both sides
    if bool(dangling_end*dangling_end2):
        gamma_current = 0.
    else:
        # check if any dangling end
        dangling_end = bool(dangling_end + dangling_end2)
        if number_of_mismatches == 0:
            alternating = bool(ii*kk*(ii!=kk)+jj*ll*(jj!=ll))
            if dangling_end:
                gamma_current = parameters['dG_3_1Match_mean']+0.5*(-1)**(1+alternating)*parameters['ddG_3_1Match_alternating']
            else:
                gamma_current = parameters['dG_4_2Match_mean']+0.5*(-1)**(1+alternating)*parameters['ddG_4_2Match_alternating']
        if number_of_mismatches == 1:
            if dangling_end:
                gamma_current = parameters['dG_3_0Match']
            else:
                gamma_current = parameters['dG_4_1Match']
        if number_of_mismatches == 2:
            gamma_current = parameters['dG_4_0Match']
    return gamma_current

def _energy_continuous_blocks(motif_shape, complements, hybridization_parameters):
    """
    ii,kk
    jj-ll
    alternating:
    X,Y Y,X
    Y-X,X-Y
    homogeneous
    X,X Y,Y
    Y-Y,X-X
    gamma_alternating <= gamma_homogeneous < gamma_one_mismatch < gamma_two_mismatches
    """
    rtrn = np.zeros(motif_shape*2)
    number_of_letters = len(complements)
    for ii,jj,kk,ll in itertools.product(range(number_of_letters+1),repeat=4):
        gamma_current = energy_continuous_block(ii,jj,kk,ll, complements, hybridization_parameters)
        #building blocks:
        #r2,r3,t4,t3
        if kk!=0 and ll !=0:
            rtrn[ii,kk-1,:,:,:,:,ll-1,jj] += gamma_current
        #r3,l2,t3,t2
        if ii*jj*kk*ll!=0:
            rtrn[:,ii-1,kk-1,:,:,ll-1,jj-1,:] += gamma_current
        #l2,l3,t2,t1
        if ii*jj!=0:
            rtrn[:,:,ii-1,kk,ll,jj-1,:,:] += gamma_current
    def make_invalid_terms_zero(rates, status_shall_be_printed=True):
        if rates.shape[1]==rates.shape[0]-1:
            rates[:,0] = 0.
            if status_shall_be_printed:
                print('Put 2nd letter to zero')
        if rates.shape[2]==rates.shape[3]-1:
            rates[:,:,0,1:]=0.
            if status_shall_be_printed:
                print('Put 3rd letter to zero, if 4th is empty')
        if rates.shape[2]==rates.shape[0]-1:
            rates[1:,:,0]=0.
            if status_shall_be_printed:
                print('Put 3rd letter to zero, if 1st is empty')
        if rates.shape[5]==rates.shape[0]-1:
            rates[:,:,:,:,:,0] = 0.
            if status_shall_be_printed:
                print('Put 6th letter to zero')
        if rates.shape[6]==rates.shape[7]-1:
            rates[:,:,:,:,:,:,0,1:] = 0.
            if status_shall_be_printed:
                print('Put 6th letter to zero, if 7th is empty')
        if rates.shape[6]==rates.shape[4]-1:
            rates[:,:,:,:,1:,:,0] = 0.
            if status_shall_be_printed:
                print('Put 7th letter to zero, if 5th is empty')
        return rates
    return rtrn

def _ligation_sites_term():
    warnings.warn('calculate_ligation_sites_term not yet implemented.')
    return 0.

def _total_hybridization_energy(motif_shape, complements, hybridization_parameters):
    the = _energy_continuous_blocks(motif_shape, complements, hybridization_parameters)
    the += _ligation_sites_term()
    return the

def _check_rotational_symmetry(number_of_letters, motif_shape):
    '''
    r2,r3,l2,l3,t1,t2,t3,t4
    r2-r3|l2-l3
    t4-t3-t2-t1
    rotational sym: (r2,r3,l2,l3)=(t1,t2,t3,t4)
    '''
    shape = motif_shape*2
    rho = np.zeros(shape)
    for r2 in range(number_of_letters+1):
        for r3 in range(number_of_letters):
            for l2 in range(number_of_letters):
                for l3 in range(number_of_letters+1):
                    rho[r2,r3,l2,l3,r2,r3,l2,l3] = 1.
    return 0.

def _calculate_entropic_term(number_of_letters, motif_shape):
    '''
    \rho \ln(2)
    \rho = 1 if strand is rotationally symmetric,
    0 else
    '''
    return _check_rotational_symmetry(number_of_letters, motif_shape) * np.log(2.)

def _total_free_energy(motif_shape,complements, hybridization_parameters, temperature, number_of_letters):
    return _total_hybridization_energy(motif_shape, complements, hybridization_parameters) + temperature*_calculate_entropic_term(number_of_letters,motif_shape)

def _calculate_stalling_factor(ii,jj,kk,ll, segment_is_to_the_left, ligation_parameters, complements):
    """
    ii,kk
    jj-ll
    alternating:
    X,Y Y,X
    Y-X,X-Y
    homogeneous
    X,X Y,Y
    Y-Y,X-X
    gamma_alternating <= gamma_homogeneous < gamma_one_mismatch < gamma_two_mismatches
    """
    sigma1 = ligation_parameters['stalling_factor_first']
    sigma2 = ligation_parameters['stalling_factor_second']
    left_mismatch = False if not ii*jj else hybridization_bond_category(ii,jj, complements)[0]
    right_mismatch = False if not kk*ll else hybridization_bond_category(kk,ll, complements)[0]
    rtrn = 1.
    if segment_is_to_the_left:
        if right_mismatch:
            rtrn *= sigma1
            if left_mismatch:
                rtrn *= sigma2
    else:
        if left_mismatch:
            rtrn *= sigma1
            if right_mismatch:
                rtrn *= sigma2
    return rtrn

def _calculate_stalling(motif_shape, number_of_letters, ligation_parameters, complements):
    stalling = np.ones(motif_shape*2)
    for ii in range(0,number_of_letters+1):
        for jj in range(0, number_of_letters+1):
            for kk in range(0, number_of_letters+1):
                for ll in range(0, number_of_letters+1):
                    segment_is_to_the_left=True
                    stalling_factor = _calculate_stalling_factor(ii,jj,kk,ll,segment_is_to_the_left, ligation_parameters, complements)
                    if kk!=0 and ll !=0:
                        stalling[ii,kk-1,:,:,:,:,ll-1,jj] *= stalling_factor
                    segment_is_to_the_left=False
                    stalling_factor = _calculate_stalling_factor(ii,jj,kk,ll,segment_is_to_the_left, ligation_parameters, complements)
                    if ii*jj!=0:
                        stalling[:,:,ii-1,kk,ll,jj-1,:,:] *= stalling_factor
    # TODO: Check, das invalid terms zero
    return stalling

def give_shape(number_of_letters):
    shape = (number_of_letters+1,number_of_letters)
    shape += shape[::-1]
    return shape



def alphabet(strand_trajectory_id : str) -> list:
    from ..read.strand_reactor_nucleotides import strand_reactor_nucleotides as read_strand_reactor_nucleotides
    from ..utils.manage_strand_reactor_files import _create_typical_alphabet_filepath
    filepath = _create_typical_alphabet_filepath(strand_trajectory_id)
    return read_strand_reactor_nucleotides(filepath)



from .effective_ligation_rates_from_parameters import effective_ligation_rates_from_parameters
from ..obj.motif_production_vector import (MotifProductionVector, _array_to_motif_production_vector)

from ..obj.units import make_unit

def motif_production_rate_constants_from_strand_reactor_parameters(
        strand_reactor_parameters : dict,
        motiflength : int,
        alphabet : list,
        maximum_ligation_window_length : int,
        complements : list,
    ) -> MotifProductionVector:
    """
    calculates (estimates for) motif_production_rate_constants
    from strand reactor parameters 
    """
    if motiflength != 4:
        raise NotImplementedError("function only implemented for fourmers (i.e. motiflength=4, given motiflength: {})".format(motiflength))
    if maximum_ligation_window_length != 4:
        raise NotImplementedError("function only implemented for maximum_ligation_window_length 4 (given maximum_ligation_window_length: {})".format(maximum_ligation_window_length))
    unit = make_unit('mol')**2/make_unit('L')**2
    motif_production_rate_constants = effective_ligation_rates_from_parameters(
        strand_reactor_parameters,
        complements = complements,
        motiflength = motiflength
    )
    return _array_to_motif_production_vector(motif_production_rate_constants, motiflength, alphabet, unit, maximum_ligation_window_length)



import numpy as np
from typing import Union
from .strand_motifs_productions_trajectory import strand_motifs_productions_trajectory as read_strand_motifs_productions_trajectory
from ..obj.motif_production_trajectory_ensemble import MotifProductionTrajectoryEnsemble
from ..obj.motif_production_trajectory import isinstance_motifproductiontrajectory, are_compatible_motif_production_trajectories
from warnings import warn

def strand_motifs_productions_trajectory_ensemble(filepath_lists : list,
        alphabet : list,
        motiflength : int,
        maximum_ligation_window_length : int,
        skiprows : int =2) -> MotifProductionTrajectoryEnsemble:
    """
    reads from the ligation_statistics.txt of the RNAReactor simulation output and returns corresponding
    production vectors in motif space.

    PARAMETERS:
    -----------
    filepath_lists : list
        Output files (ligation_statistics.txt) of the RNAReactor simulation
        Can be list of lists of motif_production_trajectory_sections, then the inner list
        will be used for creating the motif trajectories, the outer list
        distinguishes the different motif trajectories.
    skiprow : int, optional
        Skip the first `skiprow` lines, including comments
        when reading the files;
        default : 2

    RETURN:
    -------
    strand_motifs_productions_trajectory_ensemble : MotifProductionTrajectoryEnsemble
    """
    if filepath_lists is str:
        filepath_lists = [[filepath_lists,],]
    if not isinstance(filepath_lists, list):
        raise ValueError("filepaths needs to be list.")
    motifs_productions_trajectories = []

    for filepaths in filepath_lists:
        motifs_productions_trajectories = motifs_productions_trajectories + [read_strand_motifs_productions_trajectory(
            filepaths,
            alphabet,
            motiflength,
            maximum_ligation_window_length
            ),]
        assert isinstance_motifproductiontrajectory(motifs_productions_trajectories[-1]), "Not a MotifProductionTrajectory"
        if (len(motifs_productions_trajectories)>1) and not are_compatible_motif_production_trajectories(motifs_productions_trajectories[0], motifs_productions_trajectories[-1]):
            raise ValueError("Non compatible motif production trajectories.")
    return MotifProductionTrajectoryEnsemble(motifs_productions_trajectories)



import matplotlib.pyplot as plt
from matplotlib import colors
import numpy as np

def motif_production_rates(
        motif_production_rates,
        letters : list[str],
        plotpath : str = None,
        plotformats : str = ['.pdf','.pgf'],
        plot_parameters : list = [{
            'color' : 'b',
            'alpha' : 0.3,
            'label' : 'Strand',
        }],
        figname='motif_production_rates',
        ax = plt
        ):
    ax.imshow(motif_production_rates)
    cw = compute_words(letters,zeros=(1,0,0,1))
    # motif_production_rates = sort_by_length(np.array(motif_production_rates))
    # cw = sort_by_length(cw)
    ax.xticks(ticks=np.arange(0,len(cw.reshape(-1))), labels=cw.reshape(-1), fontsize='x-small', rotation=90)
    ax.yticks(ticks=np.arange(0,len(cw.reshape(-1))), labels=cw.reshape(-1), fontsize='x-small')
    if plotpath is not None:
        for plotformat in plotformats:
            plt.savefig(plotpath+figname+plotformat)
    #plot_rate_constants(motif_production_rates, 'Motif Production Rates', plotpath+'motif_production_rates.pdf')

def sort_by_length(x):
    shape = x.shape
    shape1 = shape[len(shape)//2:]
    shape2 = shape[:len(shape)//2]
    if shape1==shape2:
        if shape[1]==shape[0]:
            x = x[:,1:,1:,:,:,1:,1:,:]
            shape = x.shape
            shape1 = shape[len(shape)//2:]
            shape2 = shape[:len(shape)//2]
        dimers1 = x[0,:,:,0].reshape((-1,)+shape2)
        starts1 = x[0,:,:,1:].reshape((-1,)+shape2)
        continuations1 = x[1:,:,:,1:].reshape((-1,)+shape2)
        ends1 = x[1:,:,:,0].reshape((-1,)+shape2)
        rtrn = np.concatenate((dimers1, starts1, continuations1, ends1))
        dimers = rtrn[:,0,:,:,0].reshape((np.prod(shape1),-1))
        starts = rtrn[:,0,:,:,1:].reshape((np.prod(shape1),-1))
        continuations = rtrn[:,1:,:,:,1:].reshape((np.prod(shape1),-1))
        ends = rtrn[:,1:,:,:,0].reshape((np.prod(shape1),-1))
        rtrn = np.concatenate((dimers,starts,continuations,ends),axis=1)
    else:
        if shape[1]==shape[0]:
            x = x[:,1:,1:,:]
        dimers1 = x[0,:,:,0].reshape(-1)
        starts1 = x[0,:,:,1:].reshape(-1)
        continuations1 = x[1:,:,:,1:].reshape(-1)
        ends1 = x[1:,:,:,0].reshape(-1)
        rtrn = np.concatenate((dimers1, starts1, continuations1, ends1))
    return rtrn

def compute_words(letters,zeros=(1,0,0,1)): 
    zeros = np.asarray(zeros)
    nol = len(letters)
    shape = np.asarray((nol,)*len(zeros))
    shape = tuple(shape+zeros)
    ne = np.empty(shape, dtype=object)
    if zeros[0]:
        ne[0] = '0'
    for ii in range(nol):
        ne[ii+zeros[0]] = letters[ii]
    for jj in range(1,len(shape)):
        ne = np.swapaxes(ne,jj,0)
        if zeros[jj]:
            ne[0] += '0'
        for ii in range(nol):
            ne[ii+zeros[jj]] += letters[ii]
        ne = np.swapaxes(ne,jj,0)
    return ne

'''
def plot_rate_constants(rate_constants, title, figname,
        colorbar_norm = colors.LogNorm(),
        c_ref = trj.parameters['c_ref'],
        ticks = mean_ticks,
        ):
    fig_dct = {
            'legend' : False,
            'title' : title,
            'xlabel' : create_axis_label('template (complement)'),
            'ylabel' : create_axis_label('central produced motif'),
            'figname' : self.plotpath+figname,
            }
    plot_motifs_vs_motifs(rate_constants,
            trj.letters,
            c_ref = c_ref,
            complements = x.parameters['complements'],
            sorted_by_length = True,
            ticks = ticks,
            colorbar_norm = colorbar_norm,
            fig_dct = fig_dct
            )

def plot_motifs_vs_motifs(x,
        letters=['x','y'],
        zeros=(1,0,0,1),
        complements = None,
        c_ref = 1.,
        title=None,
        xlabel = None,
        ylabel = None,
        ticks = None,
        ax = plt,
        sorted_by_length = False,
        colorbar_norm = colors.LogNorm(),
        fig_dct = None,
        complemented_shall_be_added_to_xlabel = False
        ):
    x = transform_MultiField(x)
    if colorbar_norm is colors.LogNorm() and np.min(x)<=0:
        colorbar_norm = None
    #np.asarray(zeros+zeros)+len(letters)
    #x = x[:3,:2,:2,:3,:3,:2,:2,:3]
    if complements is not None:
        #check that complements has correct shape
        if len(complements)>len(x):
            warn("more letters given for complements than inside concentration vector. Will only take needed ones.")
        complements = complements[:len(x)]
        x = complement_template(x, complements)
        if fig_dct is not None and complemented_shall_be_added_to_xlabel:
            if fig_dct['xlabel'] is not None:
                fig_dct['xlabel'] += ' (complemented)' 
        elif complemented_shall_be_added_to_xlabel:
            if xlabel is not None:
                xlabel += ' (complemented)'
            else:
                xlabel = ' (complemented)'
    if sorted_by_length:
        x = sort_by_length(x)
        cw = sort_by_length(compute_words(letters,zeros=zeros))
    else:
        shape1 = np.prod(x.shape[:len(x.shape)//2])
        x = x.reshape((shape1,-1))
        cw = sort_by_length(compute_words(letters,zeros=zeros))
    if ticks is None:
        ticks = [None,None]
    imshow_dct = {}
    if colorbar_norm is colors.LogNorm():
        colorbar_norm = colors.LogNorm(vmin=ticks[0], vmax=ticks[1])
    else:
        imshow_dct['vmin']=ticks[0],
        imshow_dct['vmax'] = ticks[1]
    im =ax.imshow(x,
            origin='upper',
            #interpolation = 'none',
            norm = colorbar_norm,
            #FIXME: **imshow_dct
            )
    ax.colorbar(im)
    ax.xticks(ticks=np.arange(0,len(cw.reshape(-1))), labels=cw.reshape(-1), fontsize='x-small', rotation=90)
    ax.yticks(ticks=np.arange(0,len(cw.reshape(-1))), labels=cw.reshape(-1), fontsize='x-small')
    if fig_dct is not None:
        make_fig(fig_dct)
    else:
        ax.xlabel(xlabel)
        ax.ylabel(ylabel)
'''



import jax.numpy as jnp
from warnings import warn

from jax import config
config.update("jax_enable_x64", True)

def _nucleotide_mass(fourmer_log_concentration_vector):
    return jnp.sum(jnp.exp(fourmer_log_concentration_vector[:,1:]),axis=(0,2,3))+jnp.sum(jnp.exp(fourmer_log_concentration_vector[:,1:,1:,0]),axis=(0,1))

def _nucl_mass_counter(number_of_letters : int, motiflength : int):
    nucl_mass_counter = jnp.zeros((number_of_letters,)+(number_of_letters+1,)*motiflength)
    for ii in range(number_of_letters):
        nucl_mass_counter = nucl_mass_counter.at[ii,:,ii+1,:,:].add(1)
        nucl_mass_counter = nucl_mass_counter.at[ii,:,1:,ii+1,0].add(1)
    return nucl_mass_counter

def _nucleotide_mass_correction_rate(
        initial_log_concentration_array,
        current_log_concentration_array
        ):
    number_of_letters = initial_log_concentration_array.shape[0]-1
    motiflength = len(initial_log_concentration_array.shape)

    jnp.zeros((number_of_letters,)+initial_log_concentration_array.shape)
    nucl_mass_correction_rate = _nucleotide_mass(initial_log_concentration_array)-_nucleotide_mass(current_log_concentration_array)
    nucl_mass_correction_rate = nucl_mass_correction_rate.reshape((-1,)+(1,)*motiflength)

    nucl_mass_counter = _nucl_mass_counter(number_of_letters, motiflength)
    nucl_mass_correction_rate = nucl_mass_correction_rate * nucl_mass_counter * jnp.exp(current_log_concentration_array)[None]
    #return jnp.matmul(nucl_mass_counter,jnp.exp(fourmer_log_concentration_vector))

    return jnp.sum(nucl_mass_correction_rate, axis=0)

def _nonending_strand_concentration(fourmer_log_concentration_vector):
    beginning_concentration = jnp.sum(jnp.exp(fourmer_log_concentration_vector[0,1:,1:,1:]))
    ending_concentration = jnp.sum(jnp.exp(fourmer_log_concentration_vector[1:,1:,1:,0]))
    return beginning_concentration-ending_concentration

def _nonending_strand_correction_rate(
        fourmer_log_concentration_vector
        ):
    number_of_letters = fourmer_log_concentration_vector.shape[0]-1
    motiflength = len(fourmer_log_concentration_vector.shape)
    nonending_strand_concentration = jnp.zeros((number_of_letters+1,)*motiflength)
    nonending_strand_concentration = nonending_strand_concentration.at[0,1:,1:,1:].subtract(1)
    nonending_strand_concentration = nonending_strand_concentration.at[1:,1:,1:,0].add(1)
    nonending_strand_concentration = nonending_strand_concentration*jnp.exp(fourmer_log_concentration_vector)
    return _nonending_strand_concentration(fourmer_log_concentration_vector)*nonending_strand_concentration

def mass_correction_rates(
        initial_log_concentration_array : jnp.array,
        current_log_concentration_array : jnp.array,
        weight : float = 1.,
        ) -> jnp.array:
    return weight * _nonending_strand_correction_rate(current_log_concentration_array) + weight * _nucleotide_mass_correction_rate(initial_log_concentration_array, current_log_concentration_array)



import jax.numpy as jnp
from jax.experimental import sparse
from warnings import warn

def _realign_fourmer_concentration_vector(fourmer_c):
    if len(fourmer_c.shape)==1:
        # fourmer_c_format = 'slim'
        return fourmer_c
    elif fourmer_c.shape[1]==fourmer_c.shape[0]:
        fourmer_c_format = 'excessive'
    else:
        fourmer_c_format = 'extensive'
        new_fourmer_c = jnp.zeros((fourmer_c.shape[0],)*len(fourmer_c.shape))
        fourmer_c = new_fourmer_c.at[:,1:].set(fourmer_c)
    invalid_c = fourmer_c[0,0,1:,0]
    fourmer_c = fourmer_c.at[0,0,1:,0].set(fourmer_c[0,1:,0,0])
    fourmer_c = fourmer_c.at[0,1:,0,0].set(invalid_c)
    if fourmer_c_format == 'excessive':
        return fourmer_c
    else:
        return fourmer_c[:,:,1:]

def _set_invalid_logc_to_log0(
        fourmer_logc : jnp.ndarray,
        pseudo_count_concentration : float = 1.e-12,
        print_warning : bool = False,
        right_aligned : bool = False
        ) -> jnp.ndarray:
    """
    fourmer_logc : jnp.ndarray,
    pseudo_count_concentration : float = 1.e-12,
    print_warning : bool = False,
    right_aligned : bool = False
        whether strands are aligned to the left ([0,a_1,...a_l,0,...,0])
        or to the right  ([0,...,a_1,...a_l,0]),
        default is left aligned.
    """
    log0 = jnp.log(pseudo_count_concentration)
    if print_warning:
        if right_aligned:
            if jnp.any(fourmer_logc[:,:,0,:] > log0):
                warn(f"Invalid Entry in {jnp.max(fourmer_logc[:,:,0,:])=} > {log0=}")
            if jnp.any(fourmer_logc[1:,0,1:,:] > log0):
                warn(f"Invalid Entry in {jnp.max(fourmer_logc[1:,0,1:,:])=} > {log0=}")
            if jnp.any(fourmer_logc[0,0,1:,1:] > log0):
                warn(f"Invalid Entry in {jnp.max(fourmer_logc[0,0,1:,1:])=} > {log0=}")
            fourmer_logc = fourmer_logc.at[:,:,0,:].set(log0)
            fourmer_logc = fourmer_logc.at[1:,0,1:,:].set(log0)
            fourmer_logc = fourmer_logc.at[0,0,1:,1:].set(log0)
        else:
            if jnp.any(fourmer_logc[:,0,:,:] > log0):
                warn(f"Invalid Entry in {jnp.max(fourmer_logc[:,0,:,:])=} > {log0=}")
            if jnp.any(fourmer_logc[:,1:,0,1:] > log0):
                warn(f"Invalid Entry in {jnp.max(fourmer_logc[:,1:,0,1:])=} > {log0=}")
            if jnp.any(fourmer_logc[1:,1:,0,0] > log0):
                warn(f"Invalid Entry in {jnp.max(fourmer_logc[1:,1:,0,0])=} > {log0=}")
            fourmer_logc = fourmer_logc.at[:,0,:,:].set(log0)
            fourmer_logc = fourmer_logc.at[:,1:,0,1:].set(log0)
            fourmer_logc = fourmer_logc.at[1:,1:,0,0].set(log0)
    return fourmer_logc

def _set_invalid_logc_diff_to_zero(
        logc_diff : jnp.ndarray,
        print_warning : bool = True,
        right_aligned : bool = False
        ) -> jnp.ndarray:
    if print_warning:
        if right_aligned:
            if jnp.any(logc_diff[:,:,0,:] != 0.):
                warn(f"Invalid Entry in logc_diff: {jnp.max(logc_diff[:,:,0,:])=} > 0.")
            if jnp.any(logc_diff[1:,0,1:,:] != 0.):
                warn(f"Invalid Entry in logc_diff: {jnp.max(logc_diff[1:,0,1:,:])=} > 0.")
            if jnp.any(logc_diff[0,0,1:,1:] != 0.):
                warn(f"Invalid Entry in logc_diff: {jnp.max(logc_diff[0,0,1:,1:])=} > 0.")
            logc_diff = logc_diff.at[:,:,0,:].set(0.)
            logc_diff = logc_diff.at[1:,0,1:,:].set(0.)
            logc_diff = logc_diff.at[0,0,1:,1:].set(0.)
        else:
            if jnp.any(logc_diff[:,0,:,:] != 0.):
                warn(f"Invalid Entry in logc_diff: {jnp.max(logc_diff[:,0,:,:])=} > 0.")
            if jnp.any(logc_diff[:,1:,0,1:] != 0.):
                warn(f"Invalid Entry in logc_diff: {jnp.max(logc_diff[:,1:,0,1:])=} > 0.")
            if jnp.any(logc_diff[1:,1:,0,0] != 0.):
                warn(f"Invalid Entry in logc_diff: {jnp.max(logc_diff[1:,1:,0,0])=} > 0.")
        logc_diff = logc_diff.at[:,0,:,:].set(0.)
        logc_diff = logc_diff.at[:,1:,0,1:].set(0.)
        logc_diff = logc_diff.at[1:,1:,0,0].set(0.)
    return logc_diff

def _cosinus_lowcut(
        x : jnp.ndarray,
        lowcut : float = 0.,
        slope : float = 1.
        ):
    if slope == 0.:
        return jnp.where(
                x<lowcut,
                0.,
                1.)
    x = jnp.where(x<lowcut,0.,x)
    return jnp.where(
            x<lowcut+slope,
            0.5*(1-jnp.cos(jnp.pi*(x-lowcut)/slope)),
            1.)

def _cut_low_concentrations(
        logc_diff : jnp.ndarray,
        fourmer_logc : jnp.ndarray,
        soft_reactant_threshold : float = 1.e-12,
        fourmer_logc_slope : float = 0.,
        logc_diff_slope : float = 0.
        ) -> jnp.ndarray:
    if (logc_diff_slope == 0.) and (fourmer_logc_slope == 0.):
        return jnp.where((fourmer_logc<=jnp.log(soft_reactant_threshold))*(logc_diff<0.),0.,logc_diff)
    elif fourmer_logc_slope < 0.:
        raise ValueError(f"fourmer_logc_slope must be semi-positive, but\n{fourmer_logc_slope= }")
    elif logc_diff_slope< 0.:
        raise ValueError(f"logc_diff_slope must be semi-positive, but\n{logc_diff_slope = }")
    else:
        weight=1.-(1.-_cosinus_lowcut(fourmer_logc,lowcut=jnp.log(soft_reactant_threshold),slope=fourmer_logc_slope))*(1.-_cosinus_lowcut(logc_diff,lowcut=0.,slope=logc_diff_slope))
        return weight*logc_diff

def _treat_nans(
        logc_diff,
        fourmer_c,
        hard_reactant_threshold
        ) -> jnp.ndarray:
    fourmer_c[jnp.isnan(lrr)]<0.
    lrr.at[jnp.isnan(lrr)].set()
    jnp.where(jnp.isnan(log_reaction_rates))
    logc_diff = jnp.where(jnp.isnan(logc_diff)*(fourmer_c<hard_reactant_threshold),0.,logc_diff)
    if jnp.any(jnp.isnan(logc_diff)):
        warn("NaN detected detected.")
    return logc_diff

def _extract_left_reactant_c(fourmer_c):
    fourmer_c = jnp.moveaxis(fourmer_c,-1,0)
    return fourmer_c[0]

def _initizialize_left_reactant_extraction_matrix(
        numer_of_letters : int,
        motiflength : int) -> jnp.ndarray:
    # shape of left reactant
    left_reactant_shape = (jnp.sum(nol**jnp.arange(1,motiflength)),)
    # shape of full fourmer_c
    # left reactant extraction matrix
    lrem = jnp.zeros(left_reactant_shape+(numer_of_letters+1,)*motiflength)
    for strandlength in range(1,motiflength):
        if strandlength == motiflength-1:
            indcs = (slice(1,None),)*strandlength + (0,)
        else: 
            indcs = (0,) + (slice(1,None),)*strandlength + (0,)*(motiflength-strandlength-1)
        indcs_new1 = jnp.sum(nol**jnp.arange(1,strandlength))-1
        indcs_new2 = jnp.sum(nol**jnp.arange(1,strandlength+1))-1
        lrem.at[(slice(indcs_new1,indcs_new2),)+indcs].add(1)
    return lrem

def _extract_right_reactant_c(fourmer_c):
    """
    nol = fourmer_c.shape[0]-1
    motiflength = len(fourmer_c.shape)
    if fourmer_c.shape[1]+1==fourmer_c.shape[0]:
        fourmer_c = jnp.zeros((nol+1,)*motiflength).at[:,1:].set(fourmer_c)
    right_reactant_c = jnp.zeros(
            jnp.sum(nol**jnp.arange(1,motiflength))
            )
    for strandlength in range(1,motiflength):
        if strandlength == motiflength-1:
            indcs = (0,) + (slice(1,None),)*strandlength
        else: 
            indcs = (0,) + (slice(1,None),)*strandlength + (0,)*(motiflength-strandlength-1)
        indcs_new1 = jnp.sum(nol**jnp.arange(1,strandlength))-1
        indcs_new2 = jnp.sum(nol**jnp.arange(1,strandlength+1))-1
        right_reactant_c = right_reactant_c.at[indcs_new1:indcs_new2].set(fourmer_c[indcs])
    return left_reactant_c
    """
    return fourmer_c[0]

def _extract_template_c(
        fourmer_c,
        number_of_letters = None,
        ):
    """
    nol = fourmer_c.shape[0]-1
    motiflength = len(fourmer_c.shape)
    if fourmer_c.shape[1]+1==fourmer_c.shape[0]:
        fourmer_c = jnp.zeros((nol+1,)*motiflength).at[:,1:].set(fourmer_c)
    template_c = jnp.zeros(
            jnp.sum(nol**jnp.arange(2,motiflength+1)+nol**(motiflength-1))
            )
    for templatelength in range(2,motiflength+2):
        if templatelength == motiflength+1:
            indcs = (slice(1,None),)*strandlength + (0,)
        else: 
            indcs = (0,) + (slice(1,None),)*templatelength + (0,)*(motiflength-templatelength-1)
        indcs_new1 = jnp.sum(nol**jnp.arange(1,strandlength))
        indcs_new2 = jnp.sum(nol**jnp.arange(1,strandlength+1))
        right_reactant_c = right_reactant_c.at[indcs_new1:indcs_new2].set(fourmer_c[indcs])
    return template_c
    """
    if len(fourmer_c.shape)==1:
        fourmer_c_format = 'slim'
        if number_of_letters is None:
            raise ValueError("number of letters needs to be given to extract template concentration vector.")
    if len(fourmer_c.shape)!=4:
        raise NotImplementedError(f"Only implemented for motiflength 4, but {motiflength = }")
    elif fourmer_c.shape[1]==fourmer_c.shape[0]:
        fourmer_c_format = 'excessive'
        return (fourmer_c[0,1,0,1]*jnp.ones(fourmer_c.shape)).at[:,1:,1:,:].set(
                fourmer_c[:,1:,1:,:]
                )
    else:
        fourmer_c_format = 'extensive'
        return fourmer_c[:,:,1:,:]



import numpy as np
import warnings
import itertools

from .dissociation_constant import ligation_rate_constant_from_strand_reactor_parameters as infer_ligation_rate_constant_from_strand_reactor_parameters
from .dissociation_constant import dissociation_constant_from_strand_reactor_parameters as infer_dissociation_constant_from_strand_reactor_parameters

# Goeppel2022Thermodynamic Eqn A6, A30, A29
'''
according to Eqn A26
K_D = e^{\beta \Delta \mathcal G_\text{tot}(C)

β∆G_tot(C)= ∑_\text{continuous blocks} γ_i +
+ ∑_\text{dangling ends} ǫ_d +
+ ∑_\text{ligation sites} γ_l

(c^0)^2/K_{\{1,1|2\}} = \sum_{C \in \mathcal{C}_{\{1,1|2\}}} e^{-\beta
\Delta \mathcal{G}_\text{tot}(C)/|\mathcal{C}_{\{1,1|2\}}|
\lambda_{lrp} = k_\text{lig}/K_D
'''

'''
Do:5-Do:24; 8-20
Fr:9-Sa:4; Fr:10-Fr:22
Sa:13-So:8; -
So:17-Mo:12; So:22-Mo:10
Mo:21-Di:16, Di:2- Di:14
Mi:01-Mi:20, Mi:6:00 -Mi:18:00
'''

def effective_ligation_rates_from_parameters(trj_parameters : dict,
                                             complements : list = [1,0,3,2],
                                             motiflength : int = 4,
                                             standard_concentration :float = 1.
                                             ) -> np.ndarray:
    # TODO:make_differentiable
    if motiflength != 4:
        # TODO:implement_for_general_motiflength
        raise NotImplementedError("function only implemented for fourmers, yet")
    warnings.warn("This function has not been properly tested yet") # FIXME:test_function
    characteristic_ligation_length = trj_parameters['l_critical']
    hybridization_parameters, ligation_parameters = separate_hybridization_and_ligation_parameters(trj_parameters)
    return calculate_effective_ligation_rate_for_tetramer_creation(
        standard_concentration,
        characteristic_ligation_length,
        complements = complements,
        ligation_parameters = ligation_parameters,
        hybridization_parameters = hybridization_parameters
        )

def calculate_effective_ligation_rate_for_tetramer_creation(
        standard_concentration,
        characteristic_ligation_length,
        complements = [1,0,3,2],
        ligation_parameters = {'stalling_factor_first' : 1.,
            'stalling_factor_second' : 1.,
            },
        hybridization_parameters = {'dG_4_2Match_mean':0.,
                    'dG_4_1Match':0.,
                    'dG_4_0Match':0.,
                    'dG_3_1Match_mean':0.,
                    'dG_3_0Match':0.,
                    'ddG_4_2Match_alternating':0.,
                    'ddG_3_1Match_alternating':0.
                    }
        ):
    ligation_rate_constant = infer_ligation_rate_constant_from_strand_reactor_parameters(
        hybridization_parameters,
        ligation_parameters,
        complements,
        characteristic_ligation_length
    )
    # Eqn. A29
    dissociation_constant = infer_dissociation_constant_from_strand_reactor_parameters(
        characteristic_ligation_length,
        ligation_parameters,
        complements,
        hybridization_parameters
    )
    # Eqn. A30
    effective_ligation_rate_constant = ligation_rate_constant / dissociation_constant
    return effective_ligation_rate_constant

def separate_hybridization_and_ligation_parameters(
        traj_parameters : dict
    ) -> (dict,dict):
    ligation_parameters = {'stalling_factor_first' : 1.,
            'stalling_factor_second' : 1.
            }
    hybridization_parameters = {'dG_4_2Match_mean':0.,
            'dG_4_1Match':0.,
            'dG_4_0Match':0.,
            'dG_3_1Match_mean':0.,
            'dG_3_0Match':0.,
            'ddG_4_2Match_alternating':0.,
            'ddG_3_1Match_alternating':0.
            }
    for kk in ligation_parameters.keys():
        ligation_parameters[kk] = traj_parameters[kk]
    for kk in hybridization_parameters.keys():
        hybridization_parameters[kk] = traj_parameters[kk]
    return hybridization_parameters, ligation_parameters



import morsaik as kdi
import numpy as np
from numpy.testing import assert_equal
import pytest

def test_motif_trajectory():
    alphabet = ['a','b','c']
    motiflength = 5
    unit = kdi.make_unit('mol')/kdi.make_unit('L')
    times = kdi.TimesVector(np.arange(3), 'ms')

    motif_categories = kdi.domains._return_motif_categories(motiflength=motiflength)
    motif_number_dct = kdi._create_empty_motif_vector_dct(motiflength,alphabet=alphabet)

    motif_number_vector = kdi.MotifVector(motiflength,alphabet,unit)
    motif_number_vector = motif_number_vector(motif_number_dct)
    motif_number_vectors = [motif_number_vector,]*times.shape[0]
    mt = kdi.MotifTrajectory(motif_number_vectors,times)

    assert_equal(mt.times.val,times.val)
    assert(mt.times.domain[0].units,times.domain[0].units)
    assert(mt.alphabet==alphabet)
    assert(mt.number_of_letters==len(alphabet))
    assert(mt.motiflength==motiflength)
    assert(mt.unit==unit)
    for ii in range(times.shape[0]):
        for key in mt.motifs.keys():
            assert_equal(mt.motifs[key].val[ii],
                    motif_number_vectors[ii].motifs[key].val
                )

def test_motif_trajectory_assertion():
    """
    Tests that the concatentaiton of motif_vectors with different units raises AssertionError
    """
    alphabet = ['a','b','c']
    motiflength = 5
    unit = kdi.make_unit('mol')/kdi.make_unit('L')
    units2 = [kdi.make_unit(''),kdi.make_unit('mol')]
    for unit2 in units2:
        times = kdi.TimesVector(np.arange(3), 'ms')

        motif_categories = kdi.domains._return_motif_categories(motiflength=motiflength)
        motif_number_dct = kdi._create_empty_motif_vector_dct(motiflength,alphabet=alphabet)

        motif_number_vector = kdi.MotifVector(motiflength,alphabet,unit)
        motif_number_vector2 = kdi.MotifVector(motiflength,alphabet,unit2)
        motif_number_vector = motif_number_vector(motif_number_dct)
        motif_number_vector2 = motif_number_vector2(motif_number_dct)
        motif_number_vectors = [motif_number_vector,]*times.shape[0]
        motif_number_vectors[-1] = motif_number_vector2
        with pytest.raises(AssertionError):
            mt = kdi.MotifTrajectory(motif_number_vectors,times)

def test_save_and_load_motif_trajectory():
    motiflengths = [2,3,4,5,6,10]
    for motiflength in motiflengths:
        strand_trajectory_id='9999_99_99__99_99_99'
        archive_path = './archive/{strand_trajectory_id}/'.format(strand_trajectory_id=strand_trajectory_id)

        motif_trajectory = kdi.get.strand_motifs_trajectory(motiflength, strand_trajectory_id)
        kdi.save_motif_trajectory(archive_path, motif_trajectory)
        actual = kdi._motif_trajectory_as_array(
                kdi.load_motif_trajectory(archive_path)
                )
        assert_equal(actual,
                kdi._motif_trajectory_as_array(motif_trajectory)
                )

def test_array_to_motif_trajectory():
    motiflengths = [2,3,4,5,6,10]
    for motiflength in motiflengths:
        strand_trajectory_id='9999_99_99__99_99_99'
        archive_path = './archive/{strand_trajectory_id}/'.format(strand_trajectory_id=strand_trajectory_id)

        motif_trajectory = kdi.get.strand_motifs_trajectory(motiflength, strand_trajectory_id)
        motif_trajectory_array, times_array = kdi._motif_trajectory_as_array(motif_trajectory)
        actual = kdi._array_to_motif_trajectory(
                motif_trajectory_array,
                motif_trajectory.times,
                motiflength = motif_trajectory.motiflength,
                alphabet = motif_trajectory.alphabet,
                unit = motif_trajectory.unit
                )
        actual = kdi._motif_trajectory_as_array(
                actual
                )
        assert_equal(actual,
                kdi._motif_trajectory_as_array(motif_trajectory)
                )



import numpy as np
from nifty8 import MultiDomain, DomainTuple
from .hamming_space import HammingSpace
from nifty8.utilities import check_object_identity, frozendict, indent
from ..obj.units import make_unit, Unit
from warnings import warn

def _motif_categories():
    return 'length{}strand','beginning','continuation','end'

def _return_motif_categories(motiflength : int = None):
    if motiflength is None:
        return _motif_categories()
    elif motiflength < 2:
        raise("motiflength needs to be at least 2")
    elif motiflength == 2:
        return _motif_categories()[1:]
    else:
        captured_strands = tuple([_motif_categories()[0].format(ii)
            for ii in range(1,motiflength-1)])
        return captured_strands+_motif_categories()[1:]

def make_hamming_dct(alphabet, motiflength, monomers_included=True):
        hamming = lambda ii : HammingSpace(alphabet,ii)
        # domain = \sum_{k=1}^{motiflength} hamming0 * hamming**k hamming0
        # consider all strands shorter than motiflength
        monomers_untracked = 1-monomers_included
        domain = [(_return_motif_categories()[0].format(ii+1), hamming(ii+1)) for ii in range(monomers_untracked,motiflength-2)]
        # add beginnings
        domain += [(_return_motif_categories()[1], hamming(motiflength-1))]
        # add continuations
        domain += [(_return_motif_categories()[2], hamming(motiflength))]
        # add ends
        domain += [(_return_motif_categories()[3],hamming(motiflength-1))]
        domain_dct = {}
        for dom in domain:
            domain_dct[dom[0]] = dom[1]
        return domain_dct

class MotifSpace(MultiDomain):
    def __init__(self, dct,
            alphabet = None,
            motiflength = None,
            units ='bits',
            _callingfrommake=False):
        if not _callingfrommake:
            raise NotImplementedError(
                'To create a MultiDomain call `MultiDomain.make()`.')
        self._alphabet = alphabet
        self._motiflength = motiflength
        self._units = make_unit(units)
        self._keys = tuple(sorted(dct.keys()))
        self._domains = tuple(dct[key] for key in self._keys)
        self._idx = frozendict({key: i for i, key in enumerate(self._keys)})

    @staticmethod
    def make(alphabet, motiflength, units : Unit = None, monomers_included=True):
        """Creates a MultiDomain object from a dictionary of names and domains

        Parameters
        ----------
        inp : MultiDomain or dict{name: DomainTuple}
            The already built MultiDomain or a dictionary of DomainTuples

        Returns
        ------
        A MultiDomain with the input Domains as domains
        """
        if units is None:
            units = make_unit('bits',np.log2(len(alphabet)))
        else:
            units = make_unit(units)
        if (units/make_unit('bits', np.log2(len(alphabet))))!=1.:
            warn("Length of alphabet suggests that units of MotifSpace is {}".format(make_unit('bits',np.log2(len(alphabet))))+'but given units are {}.'.format(units))
        if isinstance(alphabet,int):
            alphabet = list(1 + np.arange(alphabet))
        inp = make_hamming_dct(alphabet, motiflength, monomers_included=monomers_included)

        if isinstance(inp, MotifSpace):
            return inp
        if not isinstance(inp, dict):
            raise TypeError("dict expected")
        tmp = {}
        for key, value in inp.items():
            if not isinstance(key, str):
                raise TypeError("keys must be strings")
            tmp[key] = DomainTuple.make(value)
        tmp = frozendict(tmp)
        obj = MotifSpace._domainCache.get(tmp)
        if obj is not None:
            return obj
        obj = MotifSpace(tmp,
                alphabet=alphabet,
                motiflength=motiflength,
                units=units,
                _callingfrommake=True)
        MotifSpace._domainCache[tmp] = obj
        return obj
    @property
    def motiflength(self):
        return self._motiflength
    @property
    def alphabet(self):
        return self._alphabet
    @property
    def number_of_letters(self):
        return len(self._alphabet)
    @property
    def units(self):
        return self._units



from collections import namedtuple
import numpy as np
import nifty8 as ift
import itertools
from typing import Union
import yaml
from scipy.sparse import coo_matrix, save_npz, load_npz
from os import makedirs

from warnings import warn

from .times_vector import TimesVector
from .motif_vector import isinstance_motifvector
from .motif_production_vector import (isinstance_motifproductionvector,
                                      _motif_production_array_shape,
                                      _moved_axes,
                                      _reaction_indices,
                                      _array_to_motif_production_vector,
                                      )
from .units import (Unit, make_unit,
                    transform_unit_to_dict, transform_dict_to_unit)

from ..domains.motif_space import _return_motif_categories
from ..domains.trajectory_space import TrajectorySpace
from ..domains.motif_production_space import (_determine_product_and_template_categories_and_ligation_spots,
                                              _valid_production_channel,
                                              _production_channel_id,
                                              )


from ..utils.transform_vectors_to_fields import transform_dicts_to_field


def MotifProductionTrajectory(motif_production_vectors : list,
                              times : ift.Field):
    """
    A trajectory of MotifProductionVectors.
    """
    motif_production_vector = motif_production_vectors[0]
    motif_productions = [motif_production_vectors[ii].productions for ii in range(len(motif_production_vectors))]
    motif_production_trajectory_properties = {'motiflength' : motif_production_vector.motiflength,
                                              'alphabet' : motif_production_vector.alphabet,
                                              'number_of_letters' : motif_production_vector.number_of_letters,
                                              'maximum_ligation_window_length' : motif_production_vector.maximum_ligation_window_length,
                                              'unit' : motif_production_vector.unit,
                                              'times': times,
                                              'productions' : transform_dicts_to_field(motif_productions, times.domain)
                                              }
    for motif_production_vector in motif_production_vectors:
        assert(motif_production_vector.motiflength==motif_production_trajectory_properties['motiflength'])
        assert(motif_production_vector.alphabet==motif_production_trajectory_properties['alphabet'])
        assert(motif_production_vector.number_of_letters==motif_production_trajectory_properties['number_of_letters'])
        assert motif_production_vector.unit==motif_production_trajectory_properties['unit'], "Units of motif vectors need to be the same."
    mt = namedtuple('MotifProductionTrajectory',
                tuple(motif_production_trajectory_properties.keys()))
    return mt(**motif_production_trajectory_properties)

def isinstance_motifproductiontrajectory(obj) -> bool:
    is_motif_production_trajectory = True
    keys = ['motiflength', 'alphabet', 'maximum_ligation_window_length', 'number_of_letters', 'unit', 'times', 'productions']
    for key in obj._asdict().keys():
        if key not in keys:
            print('Not a MotifProductionTrajectory, missing key: {}.'.format(key))
            return False
    is_motif_production_trajectory *= isinstance(obj, tuple)
    is_motif_production_trajectory *= hasattr(obj, '_asdict')
    is_motif_production_trajectory *= hasattr(obj, '_fields')
    return bool(is_motif_production_trajectory)

def are_compatible_motif_production_trajectories(trj1 : MotifProductionTrajectory, trj2 : MotifProductionTrajectory) -> bool:
    if not isinstance_motifproductiontrajectory(trj1):
        print('Object is not a MotifProductionTrajectory')
        return False
    if not isinstance_motifproductiontrajectory(trj2):
        print('Object is not a MotifProductionTrajectory')
        return False
    keys = ['motiflength', 'alphabet', 'maximum_ligation_window_length', 'unit']
    for key in keys:
        tt1 = trj1._asdict()[key]
        tt2 = trj2._asdict()[key]
        if isinstance(tt1, ift.Field):
            tt1 = tt1.val
        if isinstance(tt2, ift.Field):
            tt2 = tt2.val
        if not np.prod(tt1==tt2):
            print('MotifProductionTrajectories not compatible: {} mismatch.'.format(key))
            return False
    try:
        if not np.prod(trj1.times.val==trj2.times.val) or not np.prod(trj1.times.domain[0]==trj2.times.domain[0]):
            key = 'times'
            warn('MotifTrajectories {} mismatch.'.format(key))
    except ValueError:
        key = 'times'
        warn('MotifTrajectories {} mismatch.'.format(key))
    return True

def save_motif_production_trajectory(
        archive_path : str,
        motif_production_trajectory : MotifProductionTrajectory,
        file_sparse : bool = True
        ) -> None:
    makedirs(archive_path, exist_ok = True)
    productions, times = _motif_production_trajectory_as_array(motif_production_trajectory)
    if file_sparse:
        save_npz(archive_path+'motif_production_trajectory_productions',
                coo_matrix(productions.reshape(1,-1))
        )
    else:
        np.save(archive_path+'motif_production_trajectory_productions',
                productions
                )
    np.save(archive_path+'motif_production_trajectory_times',
            times
            )

    with open(archive_path+'motif_production_trajectory_properties.yaml','w') as yaml_file:
        yaml.dump({'motiflength':motif_production_trajectory.motiflength,
            'alphabet':motif_production_trajectory.alphabet,
                   'unit':transform_unit_to_dict(motif_production_trajectory.unit),
                   'times_unit' : transform_unit_to_dict(motif_production_trajectory.times.domain[0].units),
                   'maximum_ligation_window_length' : motif_production_trajectory.maximum_ligation_window_length,
                   },
            yaml_file,
            indent=4)

def load_motif_production_trajectory(
        archive_path : str,
        file_sparse : bool = True
        ) -> MotifProductionTrajectory:
    dct_filename = archive_path+'motif_production_trajectory_properties.yaml'
    productions_filename = archive_path+'motif_production_trajectory_productions'
    productions_filename += '.np'+'z'*file_sparse+'y'*(1-file_sparse)
    times_filename = archive_path+'motif_production_trajectory_times.npy'
    with open(dct_filename, 'r') as yaml_file:
        properties = yaml.safe_load(yaml_file)
    times_unit = properties['times_unit']
    del properties['times_unit']
    times_vector = TimesVector(np.load(times_filename), times_unit)
    if file_sparse:
        motif_production_trajectory_array = np.asarray(coo_matrix.todense(load_npz(productions_filename)))
        motif_production_trajectory_array = motif_production_trajectory_array.reshape(
                times_vector.domain.shape + _motif_production_array_shape(
                    len(properties["alphabet"]),
                    properties["maximum_ligation_window_length"]
                )
        )
    else:
        motif_production_trajectory_array = np.load(productions_filename)
    properties['unit'] = transform_dict_to_unit(properties['unit'])
    motif_production_trajectory = _array_to_motif_production_trajectory(times_vector, motif_production_trajectory_array,
            **properties
            )
    return motif_production_trajectory

def _array_to_motif_production_trajectory(
        times : TimesVector,
        motif_production_trajectory_array: np.ndarray,
        motiflength : int,
        alphabet : list,
        unit : Unit,
        maximum_ligation_window_length : int
        ) -> MotifProductionTrajectory:
    unit = make_unit(unit)
    motif_production_vectors = [_array_to_motif_production_vector(
        motif_production_trajectory_array[ii],
        motiflength,
        alphabet,
        unit,
        maximum_ligation_window_length
    ) for ii in range(motif_production_trajectory_array.shape[0])]
    return MotifProductionTrajectory(motif_production_vectors, times)

def _motif_production_trajectory_as_array(motif_production_trajectory : MotifProductionTrajectory
    ) -> Union[np.ndarray,np.ndarray]:
    """
    returns motifs and times of MotifTrajectory object into numpy arrays.

    Parameters:
    -----------
    motif_trajectory : MotifProducitonTrajectory

    Returns:
    --------
    motifs : np.ndarray
    times : np.ndarray
    """
    times = np.asarray(motif_production_trajectory.times.val)
    number_of_letters = motif_production_trajectory.number_of_letters
    motiflength = motif_production_trajectory.motiflength
    maximum_ligation_window_length = motif_production_trajectory.maximum_ligation_window_length

    motif_categories = _return_motif_categories(motiflength)

    motif_production_trajectory_array = np.zeros(times.shape+_motif_production_array_shape(number_of_letters,maximum_ligation_window_length))

    if maximum_ligation_window_length < 4:
        ligation_window_lengths = np.array([maximum_ligation_window_length])
    else:
        ligation_window_lengths = np.arange(4,maximum_ligation_window_length+1)

    for ligation_window_length in ligation_window_lengths:
        product_categories, template_categories, ligation_spots = _determine_product_and_template_categories_and_ligation_spots(motiflength,
                maximum_ligation_window_length,
                ligation_window_length
                )

        for product_category, template_category, ligation_spot in itertools.product(product_categories, template_categories, ligation_spots):
            if not _valid_production_channel(product_category,
                template_category,
                ligation_window_length, ligation_spot,
                maximum_ligation_window_length):
                continue
            reaction_key = _production_channel_id(product_category, template_category,
                    ligation_window_length, ligation_spot)

            destination_axes, source_axes = _moved_axes(ligation_window_length,ligation_spot, maximum_ligation_window_length)
            mpa_indices = _reaction_indices(product_category, template_category,
                    ligation_window_length, ligation_spot,
                    maximum_ligation_window_length, axes_moved=False)
            np.moveaxis(motif_production_trajectory_array,np.array(source_axes)+1,np.array(destination_axes)+1)[(slice(None),)+mpa_indices] = motif_production_trajectory.productions[reaction_key].val
    return motif_production_trajectory_array, times



import nifty8 as ift
from .motif_space import MotifSpace, _return_motif_categories
from ..obj.units import make_unit, Unit

def makeExtentiveMotifBreakageSpace(alphabet : list,
                                motiflength : int,
                                units : Unit = None
                                ):
    ms = MotifSpace.make(alphabet,
                    motiflength,
                    units
                    )
    motif_categories = _return_motif_categories(motiflength)
    ending_categories = motif_categories[:-3]+(motif_categories[-1],)
    beginning_categories = motif_categories[:-2]

    ending_space = ift.MultiDomain.make(
        {k:ms[k] for k in ending_categories if k in ms.keys()}
        # https://stackoverflow.com/a/3953386 on 202311031800
        )
    beginning_space = ift.MultiDomain.make(
        {k:ms[k] for k in beginning_categories if k in ms.keys()}
        # https://stackoverflow.com/a/3953386 on 202311031800
        )
    extentive_hyphen_space = {}
    for ending_key in ending_space.keys():
        for beginning_key in beginning_space.keys():
            hyphen_key = ending_key + '-' + beginning_key
            extentive_hyphen_space[hyphen_key] = ift.DomainTuple.make((ending_space[ending_key][0],
                                                                       beginning_space[beginning_key][0]))
    extentive_hyphen_space = ift.MultiDomain.make(extentive_hyphen_space)
    return extentive_hyphen_space



from os.path import exists

from .alphabet import alphabet as get_alphabet
from ..obj.motif_trajectory_ensemble import load_motif_trajectory_ensemble, MotifTrajectoryEnsemble
from ..read.strand_motifs_trajectory_ensemble import strand_motifs_trajectory_ensemble as read_strand_motifs_trajectory_ensemble
from ..utils.manage_strand_reactor_files import _create_typical_strand_trajectory_section_complexes_filepath 
from ..utils.save import create_trajectory_ensemble_path

def strand_motifs_trajectory_ensemble(
        motiflength : int,
        strand_trajectory_id : str,
        param_file_no : int = 0,
        skiprows : int = 2,
        execution_time_path : str = None,
        **kwargs
    ) -> MotifTrajectoryEnsemble:
    alphabet = get_alphabet(strand_trajectory_id)

    #check if already in archive
    archive_path = create_trajectory_ensemble_path(
        strand_trajectory_id=strand_trajectory_id,
        param_file_no=param_file_no,
        motiflength=motiflength
    ) + 'sd/'
    if exists(archive_path):
        return load_motif_trajectory_ensemble(archive_path)
    print('Reading MotifsTrajectoryEnsemble from strand reactor data...')
    current_filepath = lambda srn, sn : _create_typical_strand_trajectory_section_complexes_filepath(
            strand_trajectory_id,
            param_file_no=param_file_no,
            simulations_run_no=srn,
            simulations_no = sn,
            )

    simulations_run_no = 0
    simulations_no = 0
    filepath_lists = []
    while exists(current_filepath(simulations_run_no, simulations_no)):
        filepaths = []
        while exists(current_filepath(simulations_run_no, simulations_no)):
            filepaths += [current_filepath(simulations_run_no, simulations_no),]
            simulations_no += 1
        if execution_time_path is not None:
            with open(execution_time_path + 'execution_time.txt','a') as f:
                f.write(f'\nsimulations_run_no {simulations_run_no}: '+str(simulations_no))
        filepath_lists = filepath_lists + [filepaths,]
        simulations_run_no += 1
        simulations_no = 0
    return read_strand_motifs_trajectory_ensemble(filepath_lists,
            alphabet,
            motiflength,
            skiprows =skiprows)



from ..utils.manage_strand_reactor_files import _return_parameters_to_read_from_parameters_file
from json import loads, JSONDecodeError

def _read_strand_trajectory_ensemble_parameters(
        filepath : str,
        parameters_to_read_from_parameters_file : list
    ) -> dict:
    file1 = open(filepath, 'r')
    lines = file1.readlines()
    tagged_lines = (line for line in lines if any(line.startswith(tag) for tag in parameters_to_read_from_parameters_file))

    kwargs_parameters = {}
    for line in tagged_lines:
        line = line.split("//")[0]
        line = line.split("=")
        try:
            kwargs_parameters[line[0].strip()] = loads(line[1].strip())
        except JSONDecodeError:
            try:
                kwargs_parameters[line[0].strip()] = bool(line[1].strip())
            except TypeError:
                kwargs_parameters[line[0].strip()] = line[1].strip()
    c_ref = kwargs_parameters['c_ref']
    stalling = ['stalling_on', 'use_kinetic_bias_factor']
    if stalling in list(kwargs_parameters.keys()):
        if not np.prod([bool(kwargs_parameters[stall]) for stall in stalling]):
                kwargs_parameters['stalling_factor_first'] = 1.
                kwargs_parameters['stalling_factor_second'] = 1.
    #kwargs_parameters['V'] = self._motifs[0][0,1,0,0]/kwargs_parameters['c_ref']
    return kwargs_parameters

def strand_reactor_parameters(
        filepath : str,
        parameters_to_read_from_parameters_file : list = None,
    ) -> dict:
    """
    reads strand reactor parameters from strand reactor data

    Parameters:
    -----------
    filepath : str
        file path

    Returns:
    --------
    strand_reactor_parameters : dict
        Dictionary with strand reactor parameters
    """
    if parameters_to_read_from_parameters_file is None:
        parameters_to_read_from_parameters_file = _return_parameters_to_read_from_parameters_file()
    return _read_strand_trajectory_ensemble_parameters(filepath,
            parameters_to_read_from_parameters_file)



import jax.numpy as jnp
import os

def save(filename : str,
        arr : jnp.ndarray,
        allow_pickle : bool
        ):
    jnp.save(filename, arr)

def create_directory_if_not_already_existing(path):
    try:
        os.makedirs(path,mode=0o777)
        print("created directory {}".format(path))
    except(FileExistsError):
        pass

def create_directory_path_if_not_already_existing(path : str
        ) -> None:
    directories = path.split('/')
    current_path = ""
    for directory in directories:
        current_path += directory + '/'
        if directory in ('','.'):
            continue
        create_directory_if_not_already_existing(current_path)

def create_strand_trajectory_ensemble_path(
        strand_trajectory_id : str,
        param_file_no : int,
        ) -> str:
    return f'./archive/{strand_trajectory_id}/data_param_file_{param_file_no}/'

def create_trajectory_ensemble_path(
        strand_trajectory_id : str,
        param_file_no : int,
        motiflength : int,
        ) -> str:
    return create_strand_trajectory_ensemble_path(strand_trajectory_id,param_file_no)+f'{motiflength}-mer_dynamics/'


def create_motif_production_trajectory_ensemble_path(
        strand_trajectory_id : str,
        param_file_no : int,
        motiflength : int,
        maximum_ligation_window_length
        ) -> str:
    return create_trajectory_ensemble_path(strand_trajectory_id, param_file_no, motiflength) + '{}-mer_prductions/'.format(maximum_ligation_window_length)



import numpy as np
from json import loads

from ..obj.units import make_unit, Unit

def strand_length_distribution(
        filepaths : list[str],
        alphabet : list[str],
        times_unit : Unit = make_unit('t_0'),
    ) -> dict:
    rtrn = {
        'steps' : [],
        'times' : [],
        'mean_length' : [],
        'strand_length_distribution' : np.empty((0,0))
    }

    for filepath in filepaths:
        _read_length_distribution(filepath)
        steps_times_lengths, length_distribution = _read_length_distribution(filepath)
        steps, simulation_time_array, mean_length = steps_times_lengths
        rtrn['steps'] = np.concatenate((rtrn['steps'], steps))
        rtrn['times'] = np.concatenate((rtrn['times'], simulation_time_array))
        rtrn['mean_length'] = np.concatenate((rtrn['mean_length'], mean_length))

        l_max = np.max([rtrn['strand_length_distribution'].shape[1],length_distribution.shape[1]])
        sld1 = np.zeros((rtrn['strand_length_distribution'].shape[0],l_max))
        sld1[:,:rtrn['strand_length_distribution'].shape[1]] += rtrn['strand_length_distribution']
        sld2 = np.zeros((length_distribution.shape[0],l_max))
        sld2[:,:length_distribution.shape[1]] += length_distribution
        rtrn['strand_length_distribution'] = np.concatenate((sld1,sld2),axis=0)
    return rtrn

def get_overall_length_distribution(length_distribution):
    # find maximum length
    maximum_length = np.max([len(length_distribution[ii]) for ii in range(len(length_distribution))])
    overall_length_distribution = np.zeros((len(length_distribution), maximum_length), dtype=int)
    for ii in range(len(length_distribution)):
        overall_length_distribution[ii,:len(length_distribution[ii])] = np.asarray(length_distribution[ii])
    return overall_length_distribution

def _read_length_distribution(fname = None,
            keywords = ['step', 'simulation_time', '<L>'],
        ):
    def read_first_line(line,
            keywords=keywords
            ):
        for keyword in keywords:
            line = line.replace(keyword+' = ','')
        line = line.split(' ')
        for ii in range(len(line)):
            line[ii] = loads(line[ii])
        return line

    def read_strand_length_frequency(line):
        return line.split('\t')

    def make_dct_to_array(dct):
        if not isinstance(dct,dict):
            warnings.warn('Not a dictionary.')
            return dct
        lengths = [int(key) for key in dct.keys()]
        rtrn = np.zeros(np.max(lengths))
        for key in dct.keys():
            rtrn[int(key)-1] = dct[key]
        return rtrn

    f=open(fname,'r')
    lines = f.readlines()
    measurement_number = -1
    statistics = []
    for line in lines:
        line = line.replace('\n','')
        if line[:6]=='strand':
            continue
        elif line[:4]==keywords[0]:
            measurement_number += 1
            statistics.append([[],[]])
            statistics[-1][0] = read_first_line(line,
                    keywords = keywords
                    )
            statistics[-1][1] = {}
        else:
            length, ocpn = read_strand_length_frequency(line)
            statistics[-1][1][str(length)] = ocpn
    for ii in range(len(statistics)):
        statistics[ii][1] = make_dct_to_array(statistics[ii][1])
    if measurement_number == -1:
        return (np.empty(0),np.empty(0), np.empty(0)), np.empty((0,0))
    length_distribution = get_overall_length_distribution(np.asarray(statistics, dtype=object)[:,1])
    statistics = np.asarray(statistics, dtype=object)[:,0]
    steps = np.asarray([statistics[ii][0] for ii in range(len(statistics))])
    simulation_time_array = np.asarray([statistics[ii][1] for ii in range(len(statistics))])
    mean_length = np.asarray([statistics[ii][2] for ii in range(len(statistics))])
    return (steps, simulation_time_array, mean_length), length_distribution



from collections import namedtuple
import numpy as  np
from .motif_breakage_trajectory import are_compatible_motif_breakage_trajectories

def MotifBreakageTrajectoryEnsemble(motif_breakage_trajectories : list):
    """
    Ensemble of motif breakage trajectories
    """
    motif_breakage_trajectory = motif_breakage_trajectories[0]
    motif_breakage_trajectory_ensemble_properties = {
            'motiflength' : motif_breakage_trajectory.motiflength,
            'alphabet' : motif_breakage_trajectory.alphabet,
            'number_of_letters' : motif_breakage_trajectory.number_of_letters,
            'unit' : motif_breakage_trajectory.unit,
            'trajectories' : motif_breakage_trajectories
            }
    for motif_breakage_trajectory in motif_breakage_trajectories[1:]:
        assert(motif_breakage_trajectory.motiflength==motif_breakage_trajectory_ensemble_properties['motiflength'])
        assert(motif_breakage_trajectory.alphabet==motif_breakage_trajectory_ensemble_properties['alphabet'])
        assert(motif_breakage_trajectory.number_of_letters==motif_breakage_trajectory_ensemble_properties['number_of_letters'])
        assert motif_breakage_trajectory.unit==motif_breakage_trajectory_ensemble_properties['unit'], "Units of motif vectors need to be the same."
        assert(are_compatible_motif_breakage_trajectories(motif_breakage_trajectory,motif_breakage_trajectories[0]),
                "Motif trajectories not compatible.")
    mt = namedtuple('MotifBreakageTrajectoryEnsemble',
                tuple(motif_breakage_trajectory_ensemble_properties.keys()))
    return mt(**motif_breakage_trajectory_ensemble_properties)

def isinstance_motifbreakagetrajectoryensemble(obj) -> bool:
    is_motif_breakage_trajectory_ensemble = True
    keys = ['motiflength', 'alphabet', 'number_of_letters', 'unit', 'trajectories']
    for key in obj._asdict().keys():
        if key not in keys:
            print('Not a MotifProductionTrajectoryEnsemble, missing key: {}.'.format(key))
            return False
    is_motif_breakage_trajectory_ensemble *= isinstance(obj, tuple)
    is_motif_breakage_trajectory_ensemble *= hasattr(obj, '_asdict')
    is_motif_breakage_trajectory_ensemble *= hasattr(obj, '_fields')
    return bool(is_motif_breakage_trajectory_ensemble)

def are_compatible_motif_breakage_trajectory_ensembles(
        trjs1 : MotifBreakageTrajectoryEnsemble,
        trjs2 : MotifBreakageTrajectoryEnsemble
) -> bool:
    if not isinstance_motifbreakagetrajectoryensemble(trjs1):
        print('Object is not a MotifBreakageTrajectoryEnsemble')
        return False
    if not isinstance_motifbreakagetrajectoryensemble(trjs2):
        print('Object is not a MotifBreakageTrajectoryEnsemble')
        return False
    return np.prod([are_compatible_motif_breakage_trajectories(
        trj1,
        trj2
    ) for trj1 in trjs1.trajectories for trj2 in trjs2.trajectories])



import morsaik as kdi
import numpy as np
import nifty8 as ift

def test_times_vector():
    times = np.arange(128)
    times_vector = kdi.TimesVector(times, 's')
    assert(times_vector.domain==ift.DomainTuple.make(kdi.domains.TimeSpace(times,'s')))
    assert(times_vector.domain[0].units==kdi.make_unit('s'))
    times = np.copy(times)
    times[-64:]*=10
    times_vector1 = kdi.TimesVector(times,'s')
    assert(times_vector1.domain==ift.DomainTuple.make(kdi.domains.TimeSpace(times, 's')))
    assert(times_vector1.domain[0].units==kdi.make_unit('s'))
    assert(kdi.are_compatible_times_vectors(times_vector1,times_vector1))
    times_vector2 = kdi.TimesVector(times,kdi.make_unit('ms'))
    assert(not kdi.are_compatible_times_vectors(times_vector,times_vector1))
    assert(not kdi.are_compatible_times_vectors(times_vector,times_vector2))




import numpy as np
from typing import Union
from .strand_motifs_trajectory import strand_motifs_trajectory as read_strand_motifs_trajectory
from ..obj.motif_trajectory_ensemble import MotifTrajectoryEnsemble
from ..obj.motif_trajectory import isinstance_motiftrajectory, are_compatible_motif_trajectories
from warnings import warn

def strand_motifs_trajectory_ensemble(filepath_lists : list,
        alphabet : list,
        motiflength : int = 4,
        skiprows : int =2) -> MotifTrajectoryEnsemble:
    """
    reads from the complexes.txt of the RNAReactor simulation output and returns corresponding
    concentration vectors in motif space.

    PARAMETERS:
    -----------
    filepath_lists : list
        Output files (complexes.txt) of the RNAReactor simulation
        Can be list of lists of motif_trajectory_sections, then the inner list
        will be used for creating the motif trajectories, the outer list
        distinguishes the different motif trajectories.
    skiprow : int, optional
        Skip the first `skiprow` lines, including comments
        when reading the file;
        default : 2

    RETURN:
    -------
    strand_motifs_trajectory_ensemble : MotifTrajectoryEnsemble
    """
    if filepath_lists is str:
        filepath_lists = [[filepath_lists,],]
    if not isinstance(filepath_lists, list):
        raise ValueError("filepaths needs to be list.")
    motif_trajectories = []

    for filepaths in filepath_lists:
        motif_trajectories = motif_trajectories + [read_strand_motifs_trajectory(filepaths,alphabet,motiflength=motiflength),]
        assert isinstance_motiftrajectory(motif_trajectories[-1]), "Not a MotifTrajectory"
        if (len(motif_trajectories)>1) and not are_compatible_motif_trajectories(motif_trajectories[0], motif_trajectories[-1]):
            raise ValueError("Non compatible motif trajectories.")
    return MotifTrajectoryEnsemble(motif_trajectories)



from .motif_breakage_rate_constants import motif_breakage_rate_constants_from_strand_reactor_parameters

from .strand_reactor_parameters import strand_reactor_parameters
from .strand_motifs_trajectory import strand_motifs_trajectory
from .strand_motifs_trajectory_ensemble import strand_motifs_trajectory_ensemble

from .alphabet import alphabet

from .strand_motifs_productions_trajectory import strand_motifs_productions_trajectory
from .strand_motifs_productions_trajectory_ensemble import strand_motifs_productions_trajectory_ensemble

from .motif_production_rate_constants import motif_production_rate_constants_from_strand_reactor_parameters

from .strand_length_distribution import strand_length_distribution



from ..domains.motif_space import _return_motif_categories
from ..obj.units import transform_unit_to_str

import numpy as np
import itertools as it
import matplotlib.pyplot as plt
from os import makedirs

general_plot_parameters = {
        'plotpath' : None,
        'plotting_time_windows' : None,
        'md_linestyle' : '--',
        'md_color' : 'blue',
        'sd_linestyle' : '-.',
        'sd_alpha' : 0.8,
        'sd_color' : 'blue',
        }
def calculate_concentration_of_a_single_particle(c_ref,number_a_monomers):
    return c_ref/number_a_monomers

def compute_words(letters,zeros=(0,0,0,0)): 
    zeros = np.asarray(zeros)
    nol = len(letters)
    shape = np.asarray((nol,)*len(zeros))
    shape = tuple(shape+zeros)
    ne = np.empty(shape, dtype=object)
    if zeros[0]:
        ne[0] = '0'
    for ii in range(nol):
        ne[ii+zeros[0]] = letters[ii]
    for jj in range(1,len(shape)):
        ne = np.swapaxes(ne,jj,0)
        if zeros[jj]:
            ne[0] += '0'
        for ii in range(nol):
            ne[ii+zeros[jj]] += letters[ii]
        ne = np.swapaxes(ne,jj,0)
    return ne

def continuations_trajectories(
        motif_trajectory_ensembles : list,
        plotting_time_windows : list = None,
        c_ref : float = 1.,
        plotpath : str = './Plots/',
        plot_parameters : list = [{
            'linestyle' : '-',
            'color' : 'b',
            'alpha' : 0.3,
            'label' : 'Strand',
        }],
    ) -> None:
    motif_trajectories(
        motif_trajectory_ensembles,
        plotting_time_windows = plotting_time_windows,
        c_ref = c_ref,
        plotpath = plotpath,
        plot_parameters = plot_parameters,
        motifs_key = _return_motif_categories(motiflength)[-2]
    )

def motif_trajectories(
        motif_trajectory_ensembles : list,
        alphabet : list[str] = None,
        plotting_time_windows : list = None,
        c_ref : float = 1.,
        plotpath : str = './Plots/',
        plot_parameters : list = [{
            'linestyle' : '-',
            'color' : 'b',
            'alpha' : 0.3,
            'label' : 'Strand',
        }],
        plotformats : list = ['.pdf'],
        ylim : list = None,
        motifs_key : str = _return_motif_categories()[-2],
        concentration_of_a_single_particle : float = None,
        annotations : dict = {
            'X': {
                'text':'a',
                'xy':(0,1),
                'xycoords':'axes fraction',
                'xytext':(+0.5,-0.5),
                'textcoords':'offset fontsize',
                'verticalalignment':'top',
                'bbox':{
                    'facecolor':'white',
                    'edgecolor':'none',
                    'alpha':0.7,
                },
                },
            'YY':{
                'text':'b',
                'xy':(0,1),
                'xycoords':'axes fraction',
                'xytext':(+0.5,-0.5),
                'textcoords':'offset fontsize',
                'verticalalignment':'top',
                'bbox':{
                    'facecolor':'white',
                    'edgecolor':'none',
                    'alpha':0.7,
                },
                },
            'XYX':{
                'text':'c',
                'xy':(0,1),
                'xycoords':'axes fraction',
                'xytext':(+0.5,-0.5),
                'textcoords':'offset fontsize',
                'verticalalignment':'top',
                'bbox':{
                    'facecolor':'white',
                    'edgecolor':'none',
                    'alpha':0.7,
                },
                },
            'XXXX':{
                'text':'d',
                'xy':(0,1),
                'xycoords':'axes fraction',
                'xytext':(+0.5,-0.5),
                'textcoords':'offset fontsize',
                'verticalalignment':'top',
                'bbox':{
                    'facecolor':'white',
                    'edgecolor':'none',
                    'alpha':0.7,
                },
                },
            },
        **kwargs
    ) -> None:
    makedirs(plotpath, exist_ok=True)
    if not isinstance(motif_trajectory_ensembles,list):
        motif_trajectory_ensembles = [motif_trajectory_ensembles,]
    if not isinstance(plot_parameters, list):
        plot_parameters = [plot_parameters,]
    if len(plot_parameters) != len(motif_trajectory_ensembles):
        plot_parameters = plot_parameters*len(motif_trajectory_ensembles)
    motiflength = motif_trajectory_ensembles[0].motiflength
    number_of_letters = motif_trajectory_ensembles[0].number_of_letters
    if motifs_key in _return_motif_categories(motiflength)[-3:]:
        wordlength = motiflength - int(motifs_key != _return_motif_categories()[-2])
    else:
        wordlength = np.where(np.isin(_return_motif_categories(motiflength)[:-3],motifs_key))[0][0]+1
    words = compute_words(
        motif_trajectory_ensembles[0].alphabet if alphabet is None else alphabet,
        zeros=(0,)*wordlength
    )
    monomer_key = _return_motif_categories(motiflength)[0]

    motifs_list = [[motif_trajectory_ensemble.trajectories[trajectory_index] for trajectory_index in range(len(motif_trajectory_ensemble.trajectories))] for motif_trajectory_ensemble in motif_trajectory_ensembles]

    if concentration_of_a_single_particle is None:
        for ii in range(len(motif_trajectory_ensembles)):
            if isinstance(motif_trajectory_ensembles[ii].unit,(int,float)) and motif_trajectory_ensembles[ii].unit==1:
                concentration_of_a_single_particle = calculate_concentration_of_a_single_particle(c_ref, motifs_list[0][0].motifs.val[monomer_key].reshape(-1)[0])
    if concentration_of_a_single_particle is None:
        concentration_of_a_single_particle = np.min(motifs_list[0][0].motifs.val[monomer_key].reshape(-1))

    if ylim is None:
        ymax = [10**(np.ceil(np.log10(np.max(motifs[0].motifs[motifs_key].val))))*(
            1-(concentration_of_a_single_particle-1)*(isinstance(motifs[0].unit,(int,float)))
            ) for motifs in motifs_list]
        ymax = np.nanmax(list(ymax))
        if np.isnan(ymax) or np.isinf(ymax):
            ymax = None
        ylim = [concentration_of_a_single_particle/2.,ymax]

    if plotting_time_windows is None:
        plotting_time_windows = [[motif_trajectory_ensembles[0].times.val[0],motif_trajectory_ensembles[0].times.val[-1]]]

    for tptt in plotting_time_windows:
        for logt, logy in it.product(*([False,True],)*2):
            for motif_index in range(number_of_letters**wordlength):
                word = words.reshape(-1)[motif_index]
                for ii in range(len(motif_trajectory_ensembles)):
                    for motif in motifs_list[ii]:
                        if isinstance(motif_trajectory_ensembles[ii].unit,(int,float)) and motif_trajectory_ensembles[ii].unit==1:
                            mm = motif.motifs[motifs_key].val*concentration_of_a_single_particle
                        else:
                            mm = motif.motifs[motifs_key].val
                        mm = mm.reshape(len(mm),-1)
                        plt.plot(motif.times.val, mm[:,motif_index], **plot_parameters[ii])
                if word in annotations.keys():
                    plt.annotate(**annotations[word])
                plt.title(word)
                plt.xscale('linear'*(1-logt)+'log'*logt)
                plt.yscale('linear'*(1-logy)+'log'*logy)
                pth = plotpath+f'{motifs_key}_trajectories_{word}_'
                pth +=str(tptt[0])+'-'+str(tptt[1])
                pth += '_log'*logy+'_lin'*(1-logy)
                pth += '_log'*logt+'_lin'*(1-logt)
                if tptt[0] is None:
                    tptt[0] = 10**int(np.log10(motif_trajectory_ensembles[0].times.val[0]))
                if tptt[1] is None:
                    tptt[1] = np.max([motif_trajectory_ensemble.times.val[-1] for motif_trajectory_ensemble in motif_trajectory_ensembles])
                if logt and tptt[0]==0.:
                    tptt[0] = 10**int(np.log10(motif_trajectory_ensembles[0].times.val[1]))
                plt.xlim(*tptt)
                plt.ylim(ylim)
                plt.xlabel('Time ' + transform_unit_to_str(motif_trajectory_ensembles[0].times.domain[0].units))

                plt.ylabel('Concentration $[\mathrm{mol}/\mathrm{L}]$')# + transform_unit_to_str(motifs_list[0][0].unit))
                for plotformat in plotformats:
                    plt.savefig(pth+plotformat)
                print(f"Saved {pth+plotformat}.")
                plt.close('all')



import morsaik as kdi
import numpy as np
from numpy.testing import assert_equal

from morsaik import make_unit

def test_motif_number_vector():
    alphabet = ['a','b','c']
    motiflength = 5
    unit = make_unit('particles')

    motif_categories = kdi.domains._return_motif_categories(motiflength=motiflength)

    motif_number_dct = kdi._create_empty_motif_vector_dct(motiflength,alphabet=alphabet)

    motif_number_vector = kdi.MotifNumberVector(motiflength,alphabet)
    motif_number_vector = motif_number_vector(motif_number_dct)

    assert(motif_number_vector.alphabet==alphabet)
    assert(motif_number_vector.number_of_letters==len(alphabet))
    assert(motif_number_vector.motiflength==motiflength)
    assert(motif_number_vector.unit==unit)
    for motif_category in motif_categories:
        assert_equal(motif_number_vector.motifs[motif_category].val,
                motif_number_dct[motif_category].val
                )
